"""Tests covering the pipeline launch code."""

import json
from pathlib import Path
from unittest import mock

import pytest

import nf_core.pipelines.create.create
import nf_core.pipelines.launch

from ..test_pipelines import TestPipelines
from ..utils import with_temporary_file, with_temporary_folder


class TestLaunch(TestPipelines):
    def setUp(self) -> None:
        super().setUp()
        self.nf_params_fn = Path(self.pipeline_dir, "nf-params.json")
        self.launcher = nf_core.pipelines.launch.Launch(self.pipeline_dir, params_out=self.nf_params_fn)

    @mock.patch.object(nf_core.pipelines.launch.Launch, "prompt_web_gui", side_effect=[True])
    @mock.patch.object(nf_core.pipelines.launch.Launch, "launch_web_gui")
    def test_launch_pipeline(self, mock_webbrowser, mock_lauch_web_gui):
        """Test the main launch function"""
        self.launcher.launch_pipeline()

    @mock.patch.object(nf_core.pipelines.launch.Confirm, "ask", side_effect=[False])
    def test_launch_file_exists(self, mock_confirm):
        """Test that we detect an existing params file and return"""
        # Make an empty params file to be overwritten
        open(self.nf_params_fn, "a").close()

        # Try and to launch, return with error
        assert self.launcher.launch_pipeline() is False

    @mock.patch.object(nf_core.pipelines.launch.Launch, "prompt_web_gui", side_effect=[True])
    @mock.patch.object(nf_core.pipelines.launch.Launch, "launch_web_gui")
    @mock.patch.object(nf_core.pipelines.launch.Confirm, "ask", side_effect=[False])
    def test_launch_file_exists_overwrite(self, mock_webbrowser, mock_lauch_web_gui, mock_confirm):
        """Test that we detect an existing params file and we overwrite it"""
        # Make an empty params file to be overwritten
        open(self.nf_params_fn, "a").close()
        # Try and to launch, return with error
        self.launcher.launch_pipeline()

    def test_get_pipeline_schema(self):
        """Test loading the params schema from a pipeline"""
        self.launcher.get_pipeline_schema()
        assert len(self.launcher.schema_obj.schema["$defs"]["input_output_options"]["properties"]) > 2

    @with_temporary_folder
    def test_make_pipeline_schema(self, tmp_path):
        """Create a workflow, but delete the schema file, then try to load it"""
        test_pipeline_dir = Path(tmp_path, "wf")
        create_obj = nf_core.pipelines.create.create.PipelineCreate(
            "testpipeline", "a description", "Me", outdir=test_pipeline_dir, no_git=True
        )
        create_obj.init_pipeline()
        Path(test_pipeline_dir, "nextflow_schema.json").unlink()
        self.launcher = nf_core.pipelines.launch.Launch(test_pipeline_dir, params_out=self.nf_params_fn)
        self.launcher.get_pipeline_schema()
        assert len(self.launcher.schema_obj.schema["$defs"]["input_output_options"]["properties"]) >= 2
        assert self.launcher.schema_obj.schema["$defs"]["input_output_options"]["properties"]["outdir"] == {
            "type": "string",
            "format": "directory-path",
            "description": "The output directory where the results will be saved. You have to use absolute paths to storage on Cloud infrastructure.",
            "fa_icon": "fas fa-folder-open",
        }

    def test_get_pipeline_defaults(self):
        """Test fetching default inputs from the pipeline schema"""
        self.launcher.get_pipeline_schema()
        self.launcher.set_schema_inputs()
        assert len(self.launcher.schema_obj.input_params) > 0
        assert self.launcher.schema_obj.input_params["validate_params"] is True

    @with_temporary_file
    def test_get_pipeline_defaults_input_params(self, tmp_file):
        """Test fetching default inputs from the pipeline schema with an input params file supplied"""
        with open(tmp_file.name, "w") as fh:
            json.dump({"outdir": "fubar"}, fh)
        self.launcher.params_in = tmp_file.name
        self.launcher.get_pipeline_schema()
        self.launcher.set_schema_inputs()
        assert len(self.launcher.schema_obj.input_params) > 0
        assert self.launcher.schema_obj.input_params["outdir"] == "fubar"

    def test_nf_merge_schema(self):
        """Checking merging the nextflow schema with the pipeline schema"""
        self.launcher.get_pipeline_schema()
        self.launcher.set_schema_inputs()
        self.launcher.merge_nxf_flag_schema()
        assert self.launcher.schema_obj.schema["allOf"][0] == {"$ref": "#/$defs/coreNextflow"}
        assert "-resume" in self.launcher.schema_obj.schema["$defs"]["coreNextflow"]["properties"]

    def test_ob_to_questionary_string(self):
        """Check converting a python dict to a pyenquirer format - simple strings"""
        sc_obj = {
            "type": "string",
            "default": "data/*{1,2}.fastq.gz",
        }
        result = self.launcher.single_param_to_questionary("input", sc_obj)
        assert result == {
            "type": "input",
            "name": "input",
            "message": "",
            "default": "data/*{1,2}.fastq.gz",
        }

    @mock.patch("questionary.unsafe_prompt", side_effect=[{"use_web_gui": "Web based"}])
    def test_prompt_web_gui_true(self, mock_prompt):
        """Check the prompt to launch the web schema or use the cli"""
        assert self.launcher.prompt_web_gui() is True

    @mock.patch("questionary.unsafe_prompt", side_effect=[{"use_web_gui": "Command line"}])
    def test_prompt_web_gui_false(self, mock_prompt):
        """Check the prompt to launch the web schema or use the cli"""
        assert self.launcher.prompt_web_gui() is False

    @mock.patch("nf_core.utils.poll_nfcore_web_api", side_effect=[{}])
    def test_launch_web_gui_missing_keys(self, mock_poll_nfcore_web_api):
        """Check the code that opens the web browser"""
        self.launcher.get_pipeline_schema()
        self.launcher.merge_nxf_flag_schema()
        with pytest.raises(AssertionError) as exc_info:
            self.launcher.launch_web_gui()
        assert exc_info.value.args[0].startswith("Web launch response not recognised:")

    @mock.patch(
        "nf_core.utils.poll_nfcore_web_api",
        side_effect=[{"api_url": "foo", "web_url": "bar", "status": "recieved"}],
    )
    @mock.patch("webbrowser.open")
    @mock.patch("nf_core.utils.wait_cli_function")
    def test_launch_web_gui(self, mock_poll_nfcore_web_api, mock_webbrowser, mock_wait_cli_function):
        """Check the code that opens the web browser"""
        self.launcher.get_pipeline_schema()
        self.launcher.merge_nxf_flag_schema()
        assert self.launcher.launch_web_gui() is None

    @mock.patch(
        "nf_core.utils.poll_nfcore_web_api",
        side_effect=[{"status": "error", "message": "foo"}],
    )
    def test_get_web_launch_response_error(self, mock_poll_nfcore_web_api):
        """Test polling the website for a launch response - status error"""
        with pytest.raises(AssertionError) as exc_info:
            self.launcher.get_web_launch_response()
        assert exc_info.value.args[0] == "Got error from launch API (foo)"

    @mock.patch("nf_core.utils.poll_nfcore_web_api", side_effect=[{"status": "foo"}])
    def test_get_web_launch_response_unexpected(self, mock_poll_nfcore_web_api):
        """Test polling the website for a launch response - status error"""
        with pytest.raises(AssertionError) as exc_info:
            self.launcher.get_web_launch_response()
        assert exc_info.value.args[0].startswith("Web launch GUI returned unexpected status (foo): ")

    @mock.patch(
        "nf_core.utils.poll_nfcore_web_api",
        side_effect=[{"status": "waiting_for_user"}],
    )
    def test_get_web_launch_response_waiting(self, mock_poll_nfcore_web_api):
        """Test polling the website for a launch response - status waiting_for_user"""
        assert self.launcher.get_web_launch_response() is False

    @mock.patch(
        "nf_core.utils.poll_nfcore_web_api",
        side_effect=[{"status": "launch_params_complete"}],
    )
    def test_get_web_launch_response_missing_keys(self, mock_poll_nfcore_web_api):
        """Test polling the website for a launch response - complete, but missing keys"""
        with pytest.raises(AssertionError) as exc_info:
            self.launcher.get_web_launch_response()
        assert exc_info.value.args[0] == "Missing return key from web API: 'nxf_flags'"

    @mock.patch(
        "nf_core.utils.poll_nfcore_web_api",
        side_effect=[
            {
                "status": "launch_params_complete",
                "nxf_flags": {"resume", "true"},
                "input_params": {"foo", "bar"},
                "schema": {},
                "cli_launch": True,
                "nextflow_cmd": "nextflow run foo",
                "pipeline": "foo",
                "revision": "bar",
            }
        ],
    )
    @mock.patch.object(nf_core.pipelines.launch.Launch, "sanitise_web_response")
    def test_get_web_launch_response_valid(self, mock_poll_nfcore_web_api, mock_sanitise):
        """Test polling the website for a launch response - complete, valid response"""
        self.launcher.get_pipeline_schema()
        assert self.launcher.get_web_launch_response() is True

    def test_sanitise_web_response(self):
        """Check that we can properly sanitise results from the web"""
        self.launcher.get_pipeline_schema()
        self.launcher.nxf_flags["-name"] = ""
        self.launcher.schema_obj.input_params["igenomes_ignore"] = "true"
        self.launcher.sanitise_web_response()
        assert "-name" not in self.launcher.nxf_flags
        assert self.launcher.schema_obj.input_params["igenomes_ignore"] is True

    def test_ob_to_questionary_bool(self):
        """Check converting a python dict to a pyenquirer format - booleans"""
        sc_obj = {
            "type": "boolean",
            "default": "True",
        }
        result = self.launcher.single_param_to_questionary("single_end", sc_obj)
        assert result["type"] == "list"
        assert result["name"] == "single_end"
        assert result["message"] == ""
        assert result["choices"] == ["True", "False"]
        assert result["default"] == "True"
        assert result["filter"]("True") is True
        assert result["filter"]("true") is True
        assert result["filter"](True) is True
        assert result["filter"]("False") is False
        assert result["filter"]("false") is False
        assert result["filter"](False) is False

    def test_ob_to_questionary_number(self):
        """Check converting a python dict to a pyenquirer format - with enum"""
        sc_obj = {"type": "number", "default": 0.1}
        result = self.launcher.single_param_to_questionary("min_reps_consensus", sc_obj)
        assert result["type"] == "input"
        assert result["default"] == "0.1"
        assert result["validate"]("123") is True
        assert result["validate"]("-123.56") is True
        assert result["validate"]("") is True
        assert result["validate"]("123.56.78") == "Must be a number"
        assert result["validate"]("123.56sdkfjb") == "Must be a number"
        assert result["filter"]("123.456") == 123.456
        assert result["filter"]("") == ""

    def test_ob_to_questionary_integer(self):
        """Check converting a python dict to a pyenquirer format - with enum"""
        sc_obj = {"type": "integer", "default": 1}
        result = self.launcher.single_param_to_questionary("broad_cutoff", sc_obj)
        assert result["type"] == "input"
        assert result["default"] == "1"
        assert result["validate"]("123") is True
        assert result["validate"]("-123") is True
        assert result["validate"]("") is True
        assert result["validate"]("123.45") == "Must be an integer"
        assert result["validate"]("123.56sdkfjb") == "Must be an integer"
        assert result["filter"]("123") == 123
        assert result["filter"]("") == ""

    def test_ob_to_questionary_range(self):
        """Check converting a python dict to a pyenquirer format - with enum"""
        sc_obj = {"type": "number", "minimum": "10", "maximum": "20", "default": 15}
        result = self.launcher.single_param_to_questionary("broad_cutoff", sc_obj)
        assert result["type"] == "input"
        assert result["default"] == "15"
        assert result["validate"]("20") is True
        assert result["validate"]("") is True
        assert result["validate"]("123.56sdkfjb") == "Must be a number"
        assert result["validate"]("8") == "Must be greater than or equal to 10"
        assert result["validate"]("25") == "Must be less than or equal to 20"
        assert result["filter"]("20") == float(20)
        assert result["filter"]("") == ""

    def test_ob_to_questionary_enum(self):
        """Check converting a python dict to a questionary format - with enum"""
        sc_obj = {"type": "string", "default": "copy", "enum": ["symlink", "rellink"]}
        result = self.launcher.single_param_to_questionary("publish_dir_mode", sc_obj)
        assert result["type"] == "list"
        assert result["default"] == "copy"
        assert result["choices"] == ["symlink", "rellink"]

    def test_ob_to_questionary_pattern(self):
        """Check converting a python dict to a questionary format - with pattern"""
        sc_obj = {
            "type": "string",
            "pattern": "^([a-zA-Z0-9_\\-\\.]+)@([a-zA-Z0-9_\\-\\.]+)\\.([a-zA-Z]{2,5})$",
        }
        result = self.launcher.single_param_to_questionary("email", sc_obj)
        assert result["type"] == "input"
        assert result["validate"]("test@email.com") is True
        assert result["validate"]("") is True
        assert (
            result["validate"]("not_an_email")
            == r"Must match pattern: ^([a-zA-Z0-9_\-\.]+)@([a-zA-Z0-9_\-\.]+)\.([a-zA-Z]{2,5})$"
        )

    def test_strip_default_params(self):
        """Test stripping default parameters"""
        self.launcher.get_pipeline_schema()
        self.launcher.set_schema_inputs()
        self.launcher.schema_obj.input_params.update({"input": "custom_input"})
        assert len(self.launcher.schema_obj.input_params) > 1
        self.launcher.strip_default_params()
        assert self.launcher.schema_obj.input_params == {"input": "custom_input"}

    def test_build_command_empty(self):
        """Test the functionality to build a nextflow command - nothing customised"""
        self.launcher.get_pipeline_schema()
        self.launcher.merge_nxf_flag_schema()
        self.launcher.build_command()
        assert self.launcher.nextflow_cmd == f"nextflow run {self.pipeline_dir}"

    def test_build_command_nf(self):
        """Test the functionality to build a nextflow command - core nf customised"""
        self.launcher.get_pipeline_schema()
        self.launcher.merge_nxf_flag_schema()
        self.launcher.nxf_flags["-name"] = "Test_Workflow"
        self.launcher.nxf_flags["-resume"] = True
        self.launcher.build_command()
        assert self.launcher.nextflow_cmd == f'nextflow run {self.pipeline_dir} -name "Test_Workflow" -resume'

    def test_build_command_params(self):
        """Test the functionality to build a nextflow command - params supplied"""
        self.launcher.get_pipeline_schema()
        self.launcher.schema_obj.input_params.update({"input": "custom_input"})
        self.launcher.build_command()
        # Check command
        assert self.launcher.nextflow_cmd == f'nextflow run {self.pipeline_dir} -params-file "{self.nf_params_fn}"'
        # Check saved parameters file
        with open(self.nf_params_fn) as fh:
            try:
                saved_json = json.load(fh)
            except json.JSONDecodeError as e:
                raise UserWarning(f"Unable to load JSON file '{self.nf_params_fn}' due to error {e}")

        assert saved_json == {"input": "custom_input"}

    def test_build_command_params_cl(self):
        """Test the functionality to build a nextflow command - params on Nextflow command line"""
        self.launcher.use_params_file = False
        self.launcher.get_pipeline_schema()
        self.launcher.schema_obj.input_params.update({"input": "custom_input"})
        self.launcher.build_command()
        assert self.launcher.nextflow_cmd == f'nextflow run {self.pipeline_dir} --input "custom_input"'



"""
Lint the tests of a module in nf-core/modules
"""

import json
import logging
import re
from pathlib import Path

import yaml

from nf_core.components.nfcore_component import NFCoreComponent

log = logging.getLogger(__name__)


def module_tests(_, module: NFCoreComponent):
    """
    Lint the tests of a module in ``nf-core/modules``

    It verifies that the test directory exists
    and contains a ``main.nf.test`` and a ``main.nf.test.snap``

    """
    repo_dir = module.component_dir.parts[: module.component_dir.parts.index(module.component_name.split("/")[0])][-1]
    test_dir = Path(module.base_dir, "tests", "modules", repo_dir, module.component_name)
    pytest_main_nf = Path(test_dir, "main.nf")
    is_pytest = pytest_main_nf.is_file()
    if module.nftest_testdir.is_dir():
        module.passed.append(("test_dir_exists", "nf-test test directory exists", module.nftest_testdir))
    else:
        if is_pytest:
            module.warned.append(
                (
                    "test_dir_exists",
                    "nf-test directory is missing",
                    module.nftest_testdir,
                )
            )
        else:
            module.failed.append(
                (
                    "test_dir_exists",
                    "nf-test directory is missing",
                    module.nftest_testdir,
                )
            )
        return

    # Lint the test main.nf file
    if module.nftest_main_nf.is_file():
        module.passed.append(("test_main_nf_exists", "test `main.nf.test` exists", module.nftest_main_nf))
    else:
        if is_pytest:
            module.warned.append(
                (
                    "test_main_nf_exists",
                    "test `main.nf.test` does not exist",
                    module.nftest_main_nf,
                )
            )
        else:
            module.failed.append(
                (
                    "test_main_nf_exists",
                    "test `main.nf.test` does not exist",
                    module.nftest_main_nf,
                )
            )

    if module.nftest_main_nf.is_file():
        # Check if main.nf.test.snap file exists, if 'snap(' is inside main.nf.test
        with open(module.nftest_main_nf) as fh:
            if "snapshot(" in fh.read():
                snap_file = module.nftest_testdir / "main.nf.test.snap"

                if snap_file.is_file():
                    module.passed.append(
                        (
                            "test_snapshot_exists",
                            "snapshot file `main.nf.test.snap` exists",
                            snap_file,
                        )
                    )
                    # Validate no empty files
                    with open(snap_file) as snap_fh:
                        try:
                            snap_content = json.load(snap_fh)
                            for test_name in snap_content.keys():
                                if "d41d8cd98f00b204e9800998ecf8427e" in str(snap_content[test_name]):
                                    if "stub" not in test_name:
                                        module.failed.append(
                                            (
                                                "test_snap_md5sum",
                                                "md5sum for empty file found: d41d8cd98f00b204e9800998ecf8427e",
                                                snap_file,
                                            )
                                        )
                                    else:
                                        module.passed.append(
                                            (
                                                "test_snap_md5sum",
                                                "md5sum for empty file found, but it is a stub test",
                                                snap_file,
                                            )
                                        )
                                else:
                                    module.passed.append(
                                        (
                                            "test_snap_md5sum",
                                            "no md5sum for empty file found",
                                            snap_file,
                                        )
                                    )
                                if "7029066c27ac6f5ef18d660d5741979a" in str(snap_content[test_name]):
                                    if "stub" not in test_name:
                                        module.failed.append(
                                            (
                                                "test_snap_md5sum",
                                                "md5sum for compressed empty file found: 7029066c27ac6f5ef18d660d5741979a",
                                                snap_file,
                                            )
                                        )
                                    else:
                                        module.passed.append(
                                            (
                                                "test_snap_md5sum",
                                                "md5sum for compressed empty file found, but it is a stub test",
                                                snap_file,
                                            )
                                        )
                                else:
                                    module.passed.append(
                                        (
                                            "test_snap_md5sum",
                                            "no md5sum for compressed empty file found",
                                            snap_file,
                                        )
                                    )
                            if "versions" in str(snap_content[test_name]) or "versions" in str(snap_content.keys()):
                                module.passed.append(
                                    (
                                        "test_snap_versions",
                                        "versions found in snapshot file",
                                        snap_file,
                                    )
                                )
                            else:
                                module.failed.append(
                                    (
                                        "test_snap_versions",
                                        "versions not found in snapshot file",
                                        snap_file,
                                    )
                                )
                        except json.decoder.JSONDecodeError as e:
                            module.failed.append(
                                (
                                    "test_snapshot_exists",
                                    f"snapshot file `main.nf.test.snap` can't be read: {e}",
                                    snap_file,
                                )
                            )
                else:
                    module.failed.append(
                        (
                            "test_snapshot_exists",
                            "snapshot file `main.nf.test.snap` does not exist",
                            snap_file,
                        )
                    )
            # Verify that tags are correct.
            main_nf_tags = module._get_main_nf_tags(module.nftest_main_nf)
            not_alphabet = re.compile(r"[^a-zA-Z]")
            org_alp = not_alphabet.sub("", module.org)
            org_alphabet = org_alp if org_alp != "" else "nfcore"
            required_tags = ["modules", f"modules_{org_alphabet}", module.component_name]
            if module.component_name.count("/") == 1:
                required_tags.append(module.component_name.split("/")[0])
            chained_components_tags = module._get_included_components_in_chained_tests(module.nftest_main_nf)
            missing_tags = []
            log.debug(f"Required tags: {required_tags}")
            log.debug(f"Included components for chained nf-tests: {chained_components_tags}")
            for tag in set(required_tags + chained_components_tags):
                if tag not in main_nf_tags:
                    missing_tags.append(tag)
            if len(missing_tags) == 0:
                module.passed.append(
                    (
                        "test_main_tags",
                        "Tags adhere to guidelines",
                        module.nftest_main_nf,
                    )
                )
            else:
                module.failed.append(
                    (
                        "test_main_tags",
                        f"Tags do not adhere to guidelines. Tags missing in `main.nf.test`: `{','.join(missing_tags)}`",
                        module.nftest_main_nf,
                    )
                )

    # Check pytest_modules.yml does not contain entries for modules with nf-test
    pytest_yml_path = module.base_dir / "tests" / "config" / "pytest_modules.yml"
    if pytest_yml_path.is_file() and not is_pytest:
        try:
            with open(pytest_yml_path) as fh:
                pytest_yml = yaml.safe_load(fh)
                if module.component_name in pytest_yml.keys():
                    module.failed.append(
                        (
                            "test_pytest_yml",
                            "module with nf-test should not be listed in pytest_modules.yml",
                            pytest_yml_path,
                        )
                    )
                else:
                    module.passed.append(
                        (
                            "test_pytest_yml",
                            "module with  nf-test not in pytest_modules.yml",
                            pytest_yml_path,
                        )
                    )
        except FileNotFoundError:
            module.warned.append(
                (
                    "test_pytest_yml",
                    "Could not open pytest_modules.yml file",
                    pytest_yml_path,
                )
            )

    # Check that the old test directory does not exist
    if not is_pytest:
        old_test_dir = Path(module.base_dir, "tests", "modules", module.component_name)
        if old_test_dir.is_dir():
            module.failed.append(
                (
                    "test_old_test_dir",
                    f"Pytest files are still present at `{Path('tests', 'modules', module.component_name)}`. Please remove this directory and its contents.",
                    old_test_dir,
                )
            )
        else:
            module.passed.append(
                (
                    "test_old_test_dir",
                    "Old pytests don't exist for this module",
                    old_test_dir,
                )
            )



"""Linting policy for nf-core pipeline projects.

Tests Nextflow-based pipelines to check that they adhere to
the nf-core community guidelines.
"""

import datetime
import json
import logging
import os
from pathlib import Path
from typing import List, Optional, Tuple, Union

import git
import rich
import rich.progress
from git.exc import InvalidGitRepositoryError
from rich.console import group
from rich.markdown import Markdown
from rich.panel import Panel
from rich.table import Table

import nf_core.modules.lint
import nf_core.pipelines.lint_utils
import nf_core.subworkflows.lint
import nf_core.utils
from nf_core import __version__
from nf_core.components.lint import ComponentLint
from nf_core.pipelines.lint_utils import console
from nf_core.utils import NFCoreYamlConfig, NFCoreYamlLintConfig, strip_ansi_codes
from nf_core.utils import plural_s as _s

from .actions_awsfulltest import actions_awsfulltest
from .actions_awstest import actions_awstest
from .actions_ci import actions_ci
from .actions_schema_validation import actions_schema_validation
from .configs import base_config, modules_config
from .files_exist import files_exist
from .files_unchanged import files_unchanged
from .included_configs import included_configs
from .merge_markers import merge_markers
from .modules_json import modules_json
from .modules_structure import modules_structure
from .multiqc_config import multiqc_config
from .nextflow_config import nextflow_config
from .nfcore_yml import nfcore_yml
from .pipeline_name_conventions import pipeline_name_conventions
from .pipeline_todos import pipeline_todos
from .plugin_includes import plugin_includes
from .readme import readme
from .schema_description import schema_description
from .schema_lint import schema_lint
from .schema_params import schema_params
from .system_exit import system_exit
from .template_strings import template_strings
from .version_consistency import version_consistency

log = logging.getLogger(__name__)


class PipelineLint(nf_core.utils.Pipeline):
    """Object to hold linting information and results.

    Inherits :class:`nf_core.utils.Pipeline` class.

    Use the :func:`PipelineLint._lint_pipeline` function to run lint tests.

    Args:
        path (str): The path to the nf-core pipeline directory.

    Attributes:
        failed (list): A list of tuples of the form: ``(<test-name>, <reason>)``
        ignored (list): A list of tuples of the form: ``(<test-name>, <reason>)``
        lint_config (dict): The parsed nf-core linting config for this pipeline
        passed (list): A list of tuples of the form: ``(<test-name>, <reason>)``
        release_mode (bool): `True`, if you the to linting was run in release mode, `False` else.
        warned (list): A list of tuples of the form: ``(<warned no>, <reason>)``
    """

    # Import all linting tests as methods for this class
    actions_awsfulltest = actions_awsfulltest
    actions_awstest = actions_awstest
    actions_ci = actions_ci
    actions_schema_validation = actions_schema_validation
    base_config = base_config
    modules_config = modules_config
    files_exist = files_exist
    files_unchanged = files_unchanged
    merge_markers = merge_markers
    modules_json = modules_json
    modules_structure = modules_structure
    multiqc_config = multiqc_config
    nextflow_config = nextflow_config
    nfcore_yml = nfcore_yml
    pipeline_name_conventions = pipeline_name_conventions
    pipeline_todos = pipeline_todos
    plugin_includes = plugin_includes
    readme = readme
    schema_description = schema_description
    schema_lint = schema_lint
    schema_params = schema_params
    system_exit = system_exit
    template_strings = template_strings
    version_consistency = version_consistency
    included_configs = included_configs

    def __init__(
        self, wf_path, release_mode=False, fix=(), key=None, fail_ignored=False, fail_warned=False, hide_progress=False
    ):
        """Initialise linting object"""

        # Initialise the parent object
        super().__init__(wf_path)

        self.lint_config: Optional[NFCoreYamlLintConfig] = None
        self.release_mode = release_mode
        self.fail_ignored = fail_ignored
        self.fail_warned = fail_warned
        self.hide_progress = hide_progress
        self.failed = []
        self.ignored = []
        self.fixed = []
        self.passed = []
        self.warned = []
        self.could_fix = []
        self.lint_tests = self._get_all_lint_tests(self.release_mode)
        self.fix = fix
        self.key = key
        self.progress_bar = None

    @staticmethod
    def _get_all_lint_tests(release_mode):
        return [
            "files_exist",
            "nextflow_config",
            "files_unchanged",
            "actions_ci",
            "actions_awstest",
            "actions_awsfulltest",
            "readme",
            "pipeline_todos",
            "plugin_includes",
            "pipeline_name_conventions",
            "template_strings",
            "schema_lint",
            "schema_params",
            "system_exit",
            "schema_description",
            "actions_schema_validation",
            "merge_markers",
            "modules_json",
            "multiqc_config",
            "modules_structure",
            "base_config",
            "modules_config",
            "nfcore_yml",
        ] + (["version_consistency", "included_configs"] if release_mode else [])

    def _load(self) -> bool:
        """Load information about the pipeline into the PipelineLint object"""
        # Load everything using the parent object
        super()._load()

        # Load lint object specific stuff
        return self._load_lint_config()

    def _load_lint_config(self) -> bool:
        """Parse a pipeline lint config file.

        Load the '.nf-core.yml'  config file and extract
        the lint config from it

        Add parsed config to the `self.lint_config` class attribute.
        """
        _, tools_config = nf_core.utils.load_tools_config(self.wf_path)
        self.lint_config = getattr(tools_config, "lint", None) or None
        is_correct = True
        # Check if we have any keys that don't match lint test names
        if self.lint_config is not None:
            for k, v in self.lint_config:
                if v is not None and k != "nfcore_components" and k not in self.lint_tests:
                    # nfcore_components is an exception to allow custom pipelines without nf-core components
                    log.warning(f"Found unrecognised test name '{k}' in pipeline lint config")
                    is_correct = False

        return is_correct

    def _lint_pipeline(self) -> None:
        """Main linting function.

        Takes the pipeline directory as the primary input and iterates through
        the different linting checks in order. Collects any warnings or errors
        into object attributes: ``passed``, ``ignored``, ``warned`` and ``failed``.
        """
        log.info(f"Testing pipeline: [magenta]{self.wf_path}")
        if self.release_mode:
            log.info("Including --release mode tests")

        # Check that we recognise all --fix arguments
        unrecognised_fixes = list(test for test in self.fix if test not in self.lint_tests)
        if len(unrecognised_fixes):
            raise AssertionError(
                "Unrecognised lint test{} for '--fix': '{}'".format(
                    _s(unrecognised_fixes), "', '".join(unrecognised_fixes)
                )
            )

        if self.key is not None:
            # Check that supplied test keys exist
            bad_keys = [k for k in self.key if k not in self.lint_tests]
            if len(bad_keys) > 0:
                raise AssertionError(
                    "Test name{} not recognised: '{}'".format(
                        _s(bad_keys),
                        "', '".join(bad_keys),
                    )
                )

            # If -k supplied, only run these tests
            self.lint_tests = [k for k in self.lint_tests if k in self.key]
            if len(self.lint_tests) == 0:
                log.debug("No pipeline lint tests to run")
                return

        # Check that the pipeline_dir is a clean git repo
        if len(self.fix):
            log.info("Attempting to automatically fix failing tests")
            try:
                repo = git.Repo(self.wf_path)
            except InvalidGitRepositoryError:
                raise AssertionError(
                    f"'{self.wf_path}' does not appear to be a git repository, "
                    "this is required when running with '--fix'"
                )
            # Check that we have no uncommitted changes
            if repo.is_dirty(untracked_files=True):
                raise AssertionError(
                    "Uncommitted changes found in pipeline directory!\nPlease commit these before running with '--fix'"
                )

        self.progress_bar = rich.progress.Progress(
            "[bold blue]{task.description}",
            rich.progress.BarColumn(bar_width=None),
            "[magenta]{task.completed} of {task.total}[reset] » [bold yellow]{task.fields[test_name]}",
            transient=True,
            disable=self.hide_progress or os.environ.get("HIDE_PROGRESS", None) is not None,
        )
        with self.progress_bar:
            lint_progress = self.progress_bar.add_task(
                "Running lint checks", total=len(self.lint_tests), test_name=self.lint_tests[0]
            )
            for test_name in self.lint_tests:
                lint_test = self.lint_config.get(test_name, {}) if self.lint_config is not None else {}
                if lint_test is False:
                    log.debug(f"Skipping lint test '{test_name}'")
                    self.ignored.append((test_name, test_name))
                    continue
                self.progress_bar.update(lint_progress, advance=1, test_name=test_name)
                log.debug(f"Running lint test: {test_name}")
                test_results = getattr(self, test_name)()
                for test in test_results.get("passed", []):
                    self.passed.append((test_name, test))
                for test in test_results.get("ignored", []):
                    if self.fail_ignored:
                        self.failed.append((test_name, test))
                    else:
                        self.ignored.append((test_name, test))
                for test in test_results.get("fixed", []):
                    self.fixed.append((test_name, test))
                for test in test_results.get("warned", []):
                    if self.fail_warned:
                        self.failed.append((test_name, test))
                    else:
                        self.warned.append((test_name, test))
                for test in test_results.get("failed", []):
                    self.failed.append((test_name, test))
                if test_results.get("could_fix", False):
                    self.could_fix.append(test_name)

    def _print_results(self, show_passed):
        """Print linting results to the command line.

        Uses the ``rich`` library to print a set of formatted tables to the command line
        summarising the linting results.
        """

        # Spacing from log messages above
        console.print("")

        log.debug("Printing final results")

        # Helper function to format test links nicely
        @group()
        def format_result(test_results):
            """
            Given an list of error message IDs and the message texts, return a nicely formatted
            string for the terminal with appropriate ASCII colours.
            """
            tools_version = __version__
            if "dev" in __version__:
                tools_version = "latest"
            for eid, msg in test_results:
                yield Markdown(f"[{eid}](https://nf-co.re/tools/docs/{tools_version}/pipeline_lint_tests/{eid}): {msg}")

        # Table of passed tests
        if len(self.passed) > 0 and show_passed:
            console.print(
                Panel(
                    format_result(self.passed),
                    title=rf"[bold][✔] {len(self.passed)} Pipeline Test{_s(self.passed)} Passed",
                    title_align="left",
                    style="green",
                    padding=1,
                )
            )

        # Table of fixed tests
        if len(self.fixed) > 0:
            console.print(
                Panel(
                    format_result(self.fixed),
                    title=rf"[bold][?] {len(self.fixed)} Pipeline Test{_s(self.fixed)} Fixed",
                    title_align="left",
                    style="bright_blue",
                    padding=1,
                )
            )

        # Table of ignored tests
        if len(self.ignored) > 0:
            console.print(
                Panel(
                    format_result(self.ignored),
                    title=rf"[bold][?] {len(self.ignored)} Pipeline Test{_s(self.ignored)} Ignored",
                    title_align="left",
                    style="grey58",
                    padding=1,
                )
            )

        # Table of warning tests
        if len(self.warned) > 0:
            console.print(
                Panel(
                    format_result(self.warned),
                    title=rf"[bold][!] {len(self.warned)} Pipeline Test Warning{_s(self.warned)}",
                    title_align="left",
                    style="yellow",
                    padding=1,
                )
            )

        # Table of failing tests
        if len(self.failed) > 0:
            console.print(
                Panel(
                    format_result(self.failed),
                    title=rf"[bold][✗] {len(self.failed)} Pipeline Test{_s(self.failed)} Failed",
                    title_align="left",
                    style="red",
                    padding=1,
                )
            )

    def _print_summary(self):
        # Summary table
        summary_colour = "red" if len(self.failed) > 0 else "green"
        table = Table(box=rich.box.ROUNDED, style=summary_colour)
        table.add_column("LINT RESULTS SUMMARY", no_wrap=True)
        table.add_row(rf"[green][✔] {len(self.passed):>3} Test{_s(self.passed)} Passed")
        if len(self.fix):
            table.add_row(rf"[bright blue][?] {len(self.fixed):>3} Test{_s(self.fixed)} Fixed")
        table.add_row(rf"[grey58][?] {len(self.ignored):>3} Test{_s(self.ignored)} Ignored")
        table.add_row(rf"[yellow][!] {len(self.warned):>3} Test Warning{_s(self.warned)}")
        table.add_row(rf"[red][✗] {len(self.failed):>3} Test{_s(self.failed)} Failed")
        console.print(table)

    def _get_results_md(self):
        """
        Create a markdown file suitable for posting in a GitHub comment.

        Returns:
            markdown (str): Formatting markdown content
        """
        # Overall header
        overall_result = "Passed :white_check_mark:"
        if len(self.warned) > 0:
            overall_result += " :warning:"
        if len(self.failed) > 0:
            overall_result = "Failed :x:"

        tools_version = __version__
        if "dev" in __version__:
            tools_version = "latest"

        # List of tests for details
        test_failure_count = ""
        test_failures = ""
        if len(self.failed) > 0:
            test_failure_count = f"\n-| ❌ {len(self.failed):3d} tests failed       |-"
            test_failures = "### :x: Test failures:\n\n{}\n\n".format(
                "\n".join(
                    [
                        f"* [{eid}](https://nf-co.re/tools/docs/{tools_version}/pipeline_lint_tests/{eid}) - "
                        f"{strip_ansi_codes(msg, '`')}"
                        for eid, msg in self.failed
                    ]
                )
            )

        test_ignored_count = ""
        test_ignored = ""
        if len(self.ignored) > 0:
            test_ignored_count = f"\n#| ❔ {len(self.ignored):3d} tests were ignored |#"
            test_ignored = "### :grey_question: Tests ignored:\n\n{}\n\n".format(
                "\n".join(
                    [
                        f"* [{eid}](https://nf-co.re/tools/docs/{tools_version}/pipeline_lint_tests/{eid}) - "
                        f"{strip_ansi_codes(msg, '`')}"
                        for eid, msg in self.ignored
                    ]
                )
            )

        test_fixed_count = ""
        test_fixed = ""
        if len(self.fixed) > 0:
            test_fixed_count = f"\n#| ❔ {len(self.fixed):3d} tests had warnings |#"
            test_fixed = "### :grey_question: Tests fixed:\n\n{}\n\n".format(
                "\n".join(
                    [
                        f"* [{eid}](https://nf-co.re/tools/docs/{tools_version}/pipeline_lint_tests/{eid}) - "
                        f"{strip_ansi_codes(msg, '`')}"
                        for eid, msg in self.fixed
                    ]
                )
            )

        test_warning_count = ""
        test_warnings = ""
        if len(self.warned) > 0:
            test_warning_count = f"\n!| ❗ {len(self.warned):3d} tests had warnings |!"
            test_warnings = "### :heavy_exclamation_mark: Test warnings:\n\n{}\n\n".format(
                "\n".join(
                    [
                        f"* [{eid}](https://nf-co.re/tools/docs/{tools_version}/pipeline_lint_tests/{eid}) - "
                        f"{strip_ansi_codes(msg, '`')}"
                        for eid, msg in self.warned
                    ]
                )
            )

        test_passed_count = ""
        test_passes = ""
        if len(self.passed) > 0:
            test_passed_count = f"\n+| ✅ {len(self.passed):3d} tests passed       |+"
            test_passes = "### :white_check_mark: Tests passed:\n\n{}\n\n".format(
                "\n".join(
                    [
                        (
                            f"* [{eid}](https://nf-co.re/tools/docs/{tools_version}/pipeline_lint_tests/{eid})"
                            f" - {strip_ansi_codes(msg, '`')}"
                        )
                        for eid, msg in self.passed
                    ]
                )
            )

        now = datetime.datetime.now()

        comment_body_text = f"Posted for pipeline commit {self.git_sha[:7]}" if self.git_sha is not None else ""
        timestamp = now.strftime("%Y-%m-%d %H:%M:%S")
        markdown = (
            f"## `nf-core pipelines lint` overall result: {overall_result}\n\n"
            f"{comment_body_text}\n\n"
            f"```diff{test_passed_count}{test_ignored_count}{test_fixed_count}{test_warning_count}{test_failure_count}"
            "\n```\n\n"
            "<details>\n\n"
            f"{test_failures}{test_warnings}{test_ignored}{test_fixed}{test_passes}### Run details\n\n"
            f"* nf-core/tools version {nf_core.__version__}\n"
            f"* Run at `{timestamp}`\n\n"
            "</details>\n"
        )

        return markdown

    def _save_json_results(self, json_fn):
        """
        Function to dump lint results to a JSON file for downstream use

        Arguments:
            json_fn (str): File path to write JSON to.
        """

        log.info(f"Writing lint results to {json_fn}")
        now = datetime.datetime.now()
        results = {
            "nf_core_tools_version": nf_core.__version__,
            "date_run": now.strftime("%Y-%m-%d %H:%M:%S"),
            "tests_pass": [[idx, strip_ansi_codes(msg)] for idx, msg in self.passed],
            "tests_ignored": [[idx, strip_ansi_codes(msg)] for idx, msg in self.ignored],
            "tests_fixed": [[idx, strip_ansi_codes(msg)] for idx, msg in self.fixed],
            "tests_warned": [[idx, strip_ansi_codes(msg)] for idx, msg in self.warned],
            "tests_failed": [[idx, strip_ansi_codes(msg)] for idx, msg in self.failed],
            "num_tests_pass": len(self.passed),
            "num_tests_ignored": len(self.ignored),
            "num_tests_fixed": len(self.fixed),
            "num_tests_warned": len(self.warned),
            "num_tests_failed": len(self.failed),
            "has_tests_pass": len(self.passed) > 0,
            "has_tests_ignored": len(self.ignored) > 0,
            "has_tests_fixed": len(self.fixed) > 0,
            "has_tests_warned": len(self.warned) > 0,
            "has_tests_failed": len(self.failed) > 0,
            "markdown_result": self._get_results_md(),
        }
        with open(json_fn, "w") as fh:
            json.dump(results, fh, indent=4)

    def _wrap_quotes(self, files: Union[List[str], List[Path], Path]) -> str:
        """Helper function to take a list of filenames and format with markdown.

        Args:
            files (list): List of filenames, eg::

                ['foo', 'bar', 'baz']

        Returns:
            markdown (str): Formatted string of paths separated by word ``or``, eg::

                `foo` or bar` or `baz`
        """
        if not isinstance(files, list):
            files = [files]
        bfiles = [f"`{str(f)}`" for f in files]
        return " or ".join(bfiles)


def run_linting(
    pipeline_dir,
    release_mode: bool = False,
    fix=(),
    key=(),
    show_passed: bool = False,
    fail_ignored: bool = False,
    fail_warned: bool = False,
    sort_by: str = "test",
    md_fn=None,
    json_fn=None,
    hide_progress: bool = False,
) -> Tuple[PipelineLint, Optional[ComponentLint], Optional[ComponentLint]]:
    """Runs all nf-core linting checks on a given Nextflow pipeline project
    in either `release` mode or `normal` mode (default). Returns an object
    of type :class:`PipelineLint` after finished.

    Args:
        pipeline_dir (str): The path to the Nextflow pipeline root directory
        release_mode (bool): Set this to `True`, if the linting should be run in the `release` mode.
                             See :class:`PipelineLint` for more information.

    Returns:
        An object of type :class:`PipelineLint` that contains all the linting results.
        An object of type :class:`ComponentLint` that contains all the linting results for the modules.
        An object of type :class:`ComponentLint` that contains all the linting results for the subworkflows.
    """

    # Verify that the requested tests exist
    if key:
        all_tests = set(PipelineLint._get_all_lint_tests(release_mode)).union(
            set(nf_core.modules.lint.ModuleLint.get_all_module_lint_tests(is_pipeline=True))
        )
        bad_keys = [k for k in key if k not in all_tests]
        if len(bad_keys) > 0:
            raise AssertionError(
                "Test name{} not recognised: '{}'".format(
                    _s(bad_keys),
                    "', '".join(bad_keys),
                )
            )
        log.info("Only running tests: '{}'".format("', '".join(key)))

    # Check if we were given any keys, and if they match any pipeline tests
    if key:
        pipeline_keys = list(set(key).intersection(set(PipelineLint._get_all_lint_tests(release_mode))))
    else:
        # If no key is supplied, run all tests
        pipeline_keys = None

    # Create the lint object
    lint_obj = PipelineLint(pipeline_dir, release_mode, fix, pipeline_keys, fail_ignored, fail_warned, hide_progress)

    # Load the various pipeline configs
    lint_obj._load_lint_config()
    lint_obj.load_pipeline_config()

    if lint_obj.lint_config and not lint_obj.lint_config["nfcore_components"]:
        module_lint_obj = None
        subworkflow_lint_obj = None
    else:
        # Create the modules lint object
        module_lint_obj = nf_core.modules.lint.ModuleLint(pipeline_dir, hide_progress=hide_progress)
        # Create the subworkflows lint object
        try:
            subworkflow_lint_obj = nf_core.subworkflows.lint.SubworkflowLint(pipeline_dir, hide_progress=hide_progress)
        except LookupError:
            subworkflow_lint_obj = None

        # Verify that the pipeline is correctly configured and has  a modules.json file
        module_lint_obj.has_valid_directory()
        module_lint_obj.has_modules_file()
        # Run only the tests we want
        if key:
            # Select only the module lint tests
            module_lint_tests = list(
                set(key).intersection(set(nf_core.modules.lint.ModuleLint.get_all_module_lint_tests(is_pipeline=True)))
            )
            # Select only the subworkflow lint tests
            subworkflow_lint_tests = list(
                set(key).intersection(
                    set(nf_core.subworkflows.lint.SubworkflowLint.get_all_subworkflow_lint_tests(is_pipeline=True))
                )
            )
        else:
            # If no key is supplied, run the default modules tests
            module_lint_tests = list(("module_changes", "module_version"))
            subworkflow_lint_tests = list(("subworkflow_changes", "subworkflow_version"))
        module_lint_obj.filter_tests_by_key(module_lint_tests)
        if subworkflow_lint_obj is not None:
            subworkflow_lint_obj.filter_tests_by_key(subworkflow_lint_tests)

        # Set up files for component linting test
        module_lint_obj.set_up_pipeline_files()
        if subworkflow_lint_obj is not None:
            subworkflow_lint_obj.set_up_pipeline_files()

    # Run the pipeline linting tests
    try:
        lint_obj._lint_pipeline()
    except AssertionError as e:
        log.critical(f"Critical error: {e}")
        log.info("Stopping tests...")
        return lint_obj, module_lint_obj, subworkflow_lint_obj

    if module_lint_obj is not None:
        # Run the module lint tests
        if len(module_lint_obj.all_local_components) > 0:
            module_lint_obj.lint_modules(module_lint_obj.all_local_components, local=True)
        if len(module_lint_obj.all_remote_components) > 0:
            module_lint_obj.lint_modules(module_lint_obj.all_remote_components, local=False)
    if subworkflow_lint_obj is not None:
        # Run the subworkflows lint tests
        if len(subworkflow_lint_obj.all_local_components) > 0:
            subworkflow_lint_obj.lint_subworkflows(subworkflow_lint_obj.all_local_components, local=True)
        if len(subworkflow_lint_obj.all_remote_components) > 0:
            subworkflow_lint_obj.lint_subworkflows(subworkflow_lint_obj.all_remote_components, local=False)

    # Print the results
    lint_obj._print_results(show_passed)
    if module_lint_obj is not None:
        module_lint_obj._print_results(show_passed, sort_by=sort_by)
    if subworkflow_lint_obj is not None:
        subworkflow_lint_obj._print_results(show_passed, sort_by=sort_by)
    nf_core.pipelines.lint_utils.print_joint_summary(lint_obj, module_lint_obj, subworkflow_lint_obj)
    nf_core.pipelines.lint_utils.print_fixes(lint_obj)

    # Save results to Markdown file
    if md_fn is not None:
        log.info(f"Writing lint results to {md_fn}")
        markdown = lint_obj._get_results_md()
        with open(md_fn, "w") as fh:
            fh.write(markdown)

    # Save results to JSON file
    if json_fn is not None:
        lint_obj._save_json_results(json_fn)

    # Reminder about --release mode flag if we had failures
    if len(lint_obj.failed) > 0:
        if release_mode:
            log.info("Reminder: Lint tests were run in --release mode.")
    return lint_obj, module_lint_obj, subworkflow_lint_obj



"""
Verify that a subworkflow has a correct entry in the modules.json file
"""

import logging
from pathlib import Path

import nf_core
import nf_core.modules.modules_repo
import nf_core.modules.modules_utils

log = logging.getLogger(__name__)


def subworkflow_version(subworkflow_lint_object, subworkflow):
    """
    Verifies that the subworkflow has a version specified in the ``modules.json`` file

    It checks whether the subworkflow has an entry in the ``modules.json`` file
    containing a commit SHA. If that is true, it verifies that there are no
    newer version of the subworkflow available.
    """

    modules_json_path = Path(subworkflow_lint_object.directory, "modules.json")
    # Verify that a git_sha exists in the `modules.json` file for this module
    version = subworkflow_lint_object.modules_json.get_subworkflow_version(
        subworkflow.component_name, subworkflow.repo_url, subworkflow.org
    )
    if version is None:
        subworkflow.failed.append(("git_sha", "No git_sha entry in `modules.json`", modules_json_path))
        return

    subworkflow.git_sha = version
    subworkflow.passed.append(("git_sha", "Found git_sha entry in `modules.json`", modules_json_path))

    # Check whether a new version is available
    try:
        subworkflow.branch = subworkflow_lint_object.modules_json.get_component_branch(
            "subworkflows", subworkflow.component_name, subworkflow.repo_url, subworkflow.org
        )
        modules_repo = nf_core.modules.modules_repo.ModulesRepo(
            remote_url=subworkflow.repo_url, branch=subworkflow.branch
        )

        subworkflow_git_log = modules_repo.get_component_git_log(subworkflow.component_name, "subworkflows")
        if version == next(subworkflow_git_log)["git_sha"]:
            subworkflow.passed.append(
                ("subworkflow_version", "Subworkflow is in the latest version", subworkflow.component_dir)
            )
        else:
            subworkflow.warned.append(("subworkflow_version", "New version available", subworkflow.component_dir))
    except UserWarning:
        subworkflow.warned.append(("subworkflow_version", "Failed to fetch git log", subworkflow.component_dir))



import nf_core.pipelines.create.create
import nf_core.pipelines.lint

from ..test_lint import TestLint


class TestLintVersionConsistency(TestLint):
    def test_version_consistency(self):
        """Tests that config variable existence test fails with bad pipeline name"""
        new_pipeline = self._make_pipeline_copy()
        lint_obj = nf_core.pipelines.lint.PipelineLint(new_pipeline)
        lint_obj.load_pipeline_config()
        lint_obj.nextflow_config()

        result = lint_obj.version_consistency()
        assert result["passed"] == [
            "Version tags are numeric and consistent between container, release tag and config."
        ]
        assert result["failed"] == ["manifest.version was not numeric: 1.0.0dev!"]



"""
Lint the main.nf file of a module
"""

import logging
import re
import sqlite3
from pathlib import Path
from typing import List, Tuple
from urllib.parse import urlparse, urlunparse

import requests
import yaml
from rich.progress import Progress

import nf_core
import nf_core.modules.modules_utils
from nf_core.components.components_differ import ComponentsDiffer
from nf_core.components.nfcore_component import NFCoreComponent

log = logging.getLogger(__name__)


def main_nf(
    module_lint_object, module: NFCoreComponent, fix_version: bool, registry: str, progress_bar: Progress
) -> Tuple[List[str], List[str]]:
    """
    Lint a ``main.nf`` module file

    Can also be used to lint local module files,
    in which case failures will be reported as
    warnings.

    The test checks for the following:

    * Software versions and containers are valid
    * The module has a process label and it is among
      the standard ones.
    * If a ``meta`` map is defined as one of the modules
      inputs it should be defined as one of the outputs,
      and be correctly configured in the ``saveAs`` function.
    * The module script section should contain definitions
      of ``software`` and ``prefix``
    """

    inputs: List[str] = []
    outputs: List[str] = []

    # Check if we have a patch file affecting the 'main.nf' file
    # otherwise read the lines directly from the module
    lines: List[str] = []
    if module.is_patched:
        lines = ComponentsDiffer.try_apply_patch(
            module.component_type,
            module.component_name,
            module_lint_object.modules_repo.repo_path,
            module.patch_path,
            Path(module.component_dir).relative_to(module.base_dir),
            reverse=True,
        ).get("main.nf", [""])

    if len(lines) == 0:
        try:
            # Check whether file exists and load it
            with open(module.main_nf) as fh:
                lines = fh.readlines()
            module.passed.append(("main_nf_exists", "Module file exists", module.main_nf))
        except FileNotFoundError:
            module.failed.append(("main_nf_exists", "Module file does not exist", module.main_nf))
            raise FileNotFoundError(f"Module file does not exist: {module.main_nf}")

    deprecated_i = ["initOptions", "saveFiles", "getSoftwareName", "getProcessName", "publishDir"]
    if len(lines) > 0:
        lines_j = "\n".join(lines)
    else:
        lines_j = ""

    for i in deprecated_i:
        if i in lines_j:
            module.failed.append(
                (
                    "deprecated_dsl2",
                    f"`{i}` specified. No longer required for the latest nf-core/modules syntax!",
                    module.main_nf,
                )
            )

    # Go through module main.nf file and switch state according to current section
    # Perform section-specific linting
    state = "module"
    process_lines = []
    script_lines = []
    shell_lines = []
    when_lines = []
    iter_lines = iter(lines)
    for line in iter_lines:
        if re.search(r"^\s*process\s*\w*\s*{", line) and state == "module":
            state = "process"
        if re.search(r"^\s*input\s*:", line) and state in ["process"]:
            state = "input"
            continue
        if re.search(r"^\s*output\s*:", line) and state in ["input", "process"]:
            state = "output"
            continue
        if re.search(r"^\s*when\s*:", line) and state in ["input", "output", "process"]:
            state = "when"
            continue
        if re.search(r"^\s*script\s*:", line) and state in ["input", "output", "when", "process"]:
            state = "script"
            continue
        if re.search(r"^\s*shell\s*:", line) and state in ["input", "output", "when", "process"]:
            state = "shell"
            continue

        # Perform state-specific linting checks
        if state == "process" and not _is_empty(line):
            process_lines.append(line)
        if state == "input" and not _is_empty(line):
            # allow multiline tuples
            if "tuple" in line and line.count("(") <= 1:
                joint_tuple = line
                while re.sub(r"\s", "", line) != ")":
                    joint_tuple = joint_tuple + line
                    line = next(iter_lines)
                line = joint_tuple
            inputs.extend(_parse_input(module, line))
        if state == "output" and not _is_empty(line):
            outputs += _parse_output(module, line)
            outputs = list(set(outputs))  # remove duplicate 'meta's
        if state == "when" and not _is_empty(line):
            when_lines.append(line)
        if state == "script" and not _is_empty(line):
            script_lines.append(line)
        if state == "shell" and not _is_empty(line):
            shell_lines.append(line)

    # Check that we have required sections
    if not len(outputs):
        module.failed.append(("main_nf_script_outputs", "No process 'output' block found", module.main_nf))
    else:
        module.passed.append(("main_nf_script_outputs", "Process 'output' block found", module.main_nf))

    # Check the process definitions
    if check_process_section(module, process_lines, registry, fix_version, progress_bar):
        module.passed.append(("main_nf_container", "Container versions match", module.main_nf))
    else:
        module.warned.append(("main_nf_container", "Container versions do not match", module.main_nf))

    # Check the when statement
    check_when_section(module, when_lines)

    # Check that we have script or shell, not both
    if len(script_lines) and len(shell_lines):
        module.failed.append(("main_nf_script_shell", "Script and Shell found, should use only one", module.main_nf))

    # Check the script definition
    if len(script_lines):
        check_script_section(module, script_lines)

    # Check that shell uses a template
    if len(shell_lines):
        if any("template" in line for line in shell_lines):
            module.passed.append(("main_nf_shell_template", "`template` found in `shell` block", module.main_nf))
        else:
            module.failed.append(("main_nf_shell_template", "No `template` found in `shell` block", module.main_nf))

    # Check whether 'meta' is emitted when given as input
    if inputs:
        if "meta" in inputs:
            module.has_meta = True
            if outputs:
                if "meta" in outputs:
                    module.passed.append(
                        ("main_nf_meta_output", "'meta' map emitted in output channel(s)", module.main_nf)
                    )
                else:
                    module.failed.append(
                        ("main_nf_meta_output", "'meta' map not emitted in output channel(s)", module.main_nf)
                    )

    # Check that a software version is emitted
    if outputs:
        if "versions" in outputs:
            module.passed.append(("main_nf_version_emitted", "Module emits software version", module.main_nf))
        else:
            module.warned.append(("main_nf_version_emitted", "Module does not emit software version", module.main_nf))

    return inputs, outputs


def check_script_section(self, lines):
    """
    Lint the script section
    Checks whether `def prefix` is defined and whether getProcessName is used for `versions.yml`.
    """
    script = "".join(lines)

    # check that process name is used for `versions.yml`
    if re.search(r"\$\{\s*task\.process\s*\}", script):
        self.passed.append(("main_nf_version_script", "Process name used for versions.yml", self.main_nf))
    else:
        self.warned.append(("main_nf_version_script", "Process name not used for versions.yml", self.main_nf))

    # check for prefix (only if module has a meta map as input)
    if self.has_meta:
        if re.search(r"\s*prefix\s*=\s*task.ext.prefix", script):
            self.passed.append(("main_nf_meta_prefix", "'prefix' specified in script section", self.main_nf))
        else:
            self.failed.append(("main_nf_meta_prefix", "'prefix' unspecified in script section", self.main_nf))


def check_when_section(self, lines):
    """
    Lint the when: section
    Checks whether the line is modified from 'task.ext.when == null || task.ext.when'
    """
    if len(lines) == 0:
        self.failed.append(("when_exist", "when: condition has been removed", self.main_nf))
        return
    if len(lines) > 1:
        self.failed.append(("when_exist", "when: condition has too many lines", self.main_nf))
        return
    self.passed.append(("when_exist", "when: condition is present", self.main_nf))

    # Check the condition hasn't been changed.
    if lines[0].strip() != "task.ext.when == null || task.ext.when":
        self.failed.append(("when_condition", "when: condition has been altered", self.main_nf))
        return
    self.passed.append(("when_condition", "when: condition is unchanged", self.main_nf))


def check_process_section(self, lines, registry, fix_version, progress_bar):
    """Lint the section of a module between the process definition
    and the 'input:' definition
    Specifically checks for correct software versions
    and containers

    Args:
        lines (List[str]): Content of process.
        registry (str): Base Docker registry for containers. Typically quay.io.
        fix_version (bool): Fix software version
        progress_bar (ProgressBar): Progress bar to update.

    Returns:
        Optional[bool]: True if singularity and docker containers match, False otherwise. If process definition does not exist, None.
    """
    # Check that we have a process section
    if len(lines) == 0:
        self.failed.append(("process_exist", "Process definition does not exist", self.main_nf))
        return
    self.passed.append(("process_exist", "Process definition exists", self.main_nf))

    # Checks that build numbers of bioconda, singularity and docker container are matching
    singularity_tag = None
    docker_tag = None
    bioconda_packages = []

    # Process name should be all capital letters
    if all(x.upper() for x in self.process_name):
        self.passed.append(("process_capitals", "Process name is in capital letters", self.main_nf))
    else:
        self.failed.append(("process_capitals", "Process name is not in capital letters", self.main_nf))

    # Check that process labels are correct
    check_process_labels(self, lines)

    # Deprecated enable_conda
    for i, raw_line in enumerate(lines):
        url = None
        line = raw_line.strip(" \n'\"}:")

        # Catch preceding "container "
        if line.startswith("container"):
            line = line.replace("container", "").strip(" \n'\"}:")

        if _container_type(line) == "conda":
            if "bioconda::" in line:
                bioconda_packages = [b for b in line.split() if "bioconda::" in b]
            match = re.search(r"params\.enable_conda", line)
            if match is None:
                self.passed.append(
                    (
                        "deprecated_enable_conda",
                        "Deprecated parameter 'params.enable_conda' correctly not found in the conda definition",
                        self.main_nf,
                    )
                )
            else:
                self.failed.append(
                    (
                        "deprecated_enable_conda",
                        "Found deprecated parameter 'params.enable_conda' in the conda definition",
                        self.main_nf,
                    )
                )
        if _container_type(line) == "singularity":
            # e.g. "https://containers.biocontainers.pro/s3/SingImgsRepo/biocontainers/v1.2.0_cv1/biocontainers_v1.2.0_cv1.img -> v1.2.0_cv1
            # e.g. "https://depot.galaxyproject.org/singularity/fastqc:0.11.9--0 -> 0.11.9--0
            # Please god let's find a better way to do this than regex
            match = re.search(r"(?:[:.])?([A-Za-z\d\-_.]+?)(?:\.img)?(?:\.sif)?$", line)
            if match is not None:
                singularity_tag = match.group(1)
                self.passed.append(("singularity_tag", f"Found singularity tag: {singularity_tag}", self.main_nf))
            else:
                self.failed.append(("singularity_tag", "Unable to parse singularity tag", self.main_nf))
                singularity_tag = None
            url = urlparse(line.split("'")[0])

        if _container_type(line) == "docker":
            # e.g. "quay.io/biocontainers/krona:2.7.1--pl526_5 -> 2.7.1--pl526_5
            # e.g. "biocontainers/biocontainers:v1.2.0_cv1 -> v1.2.0_cv1
            match = re.search(r":([A-Za-z\d\-_.]+)$", line)
            if match is not None:
                docker_tag = match.group(1)
                self.passed.append(("docker_tag", f"Found docker tag: {docker_tag}", self.main_nf))
            else:
                self.failed.append(("docker_tag", "Unable to parse docker tag", self.main_nf))
                docker_tag = None
            if line.startswith(registry):
                l_stripped = re.sub(r"\W+$", "", line)
                self.failed.append(
                    (
                        "container_links",
                        f"{l_stripped} container name found, please use just 'organisation/container:tag' instead.",
                        self.main_nf,
                    )
                )
            else:
                self.passed.append(("container_links", "Container prefix is correct", self.main_nf))

            # Guess if container name is simple one (e.g. nfcore/ubuntu:20.04)
            # If so, add quay.io as default container prefix
            if line.count("/") == 1 and line.count(":") == 1:
                line = "/".join([registry, line]).replace("//", "/")
            url = urlparse(line.split("'")[0])

        if line.startswith("container") or _container_type(line) == "docker" or _container_type(line) == "singularity":
            check_container_link_line(self, raw_line, registry)

        # Try to connect to container URLs
        if url is None:
            continue
        try:
            container_url = "https://" + urlunparse(url) if not url.scheme == "https" else urlunparse(url)
            log.debug(f"Trying to connect to URL: {container_url}")
            response = requests.head(
                container_url,
                stream=True,
                allow_redirects=True,
            )
            log.debug(
                f"Connected to URL: {'https://' + urlunparse(url) if not url.scheme == 'https' else urlunparse(url)}, "
                f"status_code: {response.status_code}"
            )
        except (requests.exceptions.RequestException, sqlite3.InterfaceError) as e:
            log.debug(f"Unable to connect to url '{urlunparse(url)}' due to error: {e}")
            self.failed.append(("container_links", "Unable to connect to container URL", self.main_nf))
            continue
        if not response.ok:
            self.warned.append(
                (
                    "container_links",
                    f"Unable to connect to container registry, code:  {response.status_code}, url: {response.url}",
                    self.main_nf,
                )
            )

    # Get bioconda packages from environment.yml
    try:
        with open(Path(self.component_dir, "environment.yml")) as fh:
            env_yml = yaml.safe_load(fh)
        if "dependencies" in env_yml:
            bioconda_packages = [x for x in env_yml["dependencies"] if isinstance(x, str) and "bioconda::" in x]
    except FileNotFoundError:
        pass
    except NotADirectoryError:
        pass

    # Check that all bioconda packages have build numbers
    # Also check for newer versions
    for bp in bioconda_packages:
        bp = bp.strip("'").strip('"')
        # Check for correct version and newer versions
        try:
            bioconda_version = bp.split("=")[1]
            # response = _bioconda_package(bp)
            response = nf_core.utils.anaconda_package(bp)
        except LookupError:
            self.warned.append(("bioconda_version", f"Conda version not specified correctly: {bp}", self.main_nf))
        except ValueError:
            self.failed.append(("bioconda_version", f"Conda version not specified correctly: {bp}", self.main_nf))
        else:
            # Check that required version is available at all
            if bioconda_version not in response.get("versions"):
                self.failed.append(
                    ("bioconda_version", f"Conda package {bp} had unknown version: `{bioconda_version}`", self.main_nf)
                )
                continue  # No need to test for latest version, continue linting
            # Check version is latest available
            last_ver = response.get("latest_version")
            if last_ver is not None and last_ver != bioconda_version:
                package, ver = bp.split("=", 1)
                # If a new version is available and fix is True, update the version
                if fix_version:
                    try:
                        fixed = _fix_module_version(self, bioconda_version, last_ver, singularity_tag, response)
                    except FileNotFoundError as e:
                        fixed = False
                        log.debug(f"Unable to update package {package} due to error: {e}")
                    else:
                        if fixed:
                            progress_bar.print(f"[blue]INFO[/blue]\t Updating package '{package}' {ver} -> {last_ver}")
                            log.debug(f"Updating package {package} {ver} -> {last_ver}")
                            self.passed.append(
                                (
                                    "bioconda_latest",
                                    f"Conda package has been updated to the latest available: `{bp}`",
                                    self.main_nf,
                                )
                            )
                        else:
                            progress_bar.print(
                                f"[blue]INFO[/blue]\t Tried to update package. Unable to update package '{package}' {ver} -> {last_ver}"
                            )
                            log.debug(f"Unable to update package {package} {ver} -> {last_ver}")
                            self.warned.append(
                                ("bioconda_latest", f"Conda update: {package} `{ver}` -> `{last_ver}`", self.main_nf)
                            )
                # Add available update as a warning
                else:
                    self.warned.append(
                        ("bioconda_latest", f"Conda update: {package} `{ver}` -> `{last_ver}`", self.main_nf)
                    )
            else:
                self.passed.append(("bioconda_latest", f"Conda package is the latest available: `{bp}`", self.main_nf))

    # Check if a tag exists at all. If not, return None.
    if singularity_tag is None or docker_tag is None:
        return None
    else:
        return docker_tag == singularity_tag


def check_process_labels(self, lines):
    correct_process_labels = [
        "process_single",
        "process_low",
        "process_medium",
        "process_high",
        "process_long",
        "process_high_memory",
    ]
    all_labels = [line.strip() for line in lines if line.lstrip().startswith("label ")]
    bad_labels = []
    good_labels = []
    if len(all_labels) > 0:
        for label in all_labels:
            try:
                label = re.match(r"^label\s+'?\"?([a-zA-Z0-9_-]+)'?\"?$", label).group(1)
            except AttributeError:
                self.warned.append(
                    (
                        "process_standard_label",
                        f"Specified label appears to contain non-alphanumerics: {label}",
                        self.main_nf,
                    )
                )
                continue
            if label not in correct_process_labels:
                bad_labels.append(label)
            else:
                good_labels.append(label)
        if len(good_labels) > 1:
            self.warned.append(
                (
                    "process_standard_label",
                    f"Conflicting process labels found: `{'`,`'.join(good_labels)}`",
                    self.main_nf,
                )
            )
        elif len(good_labels) == 1:
            self.passed.append(("process_standard_label", "Correct process label", self.main_nf))
        else:
            self.warned.append(("process_standard_label", "Standard process label not found", self.main_nf))
        if len(bad_labels) > 0:
            self.warned.append(
                ("process_standard_label", f"Non-standard labels found: `{'`,`'.join(bad_labels)}`", self.main_nf)
            )
        if len(all_labels) > len(set(all_labels)):
            self.warned.append(
                (
                    "process_standard_label",
                    f"Duplicate labels found: `{'`,`'.join(sorted(all_labels))}`",
                    self.main_nf,
                )
            )
    else:
        self.warned.append(("process_standard_label", "Process label not specified", self.main_nf))


def check_container_link_line(self, raw_line, registry):
    """Look for common problems in the container name / URL, for docker and singularity."""

    line = raw_line.strip(" \n'\"}:")

    # lint double quotes
    if line.count('"') > 2:
        self.failed.append(
            (
                "container_links",
                f"Too many double quotes found when specifying container: {line.lstrip('container ')}",
                self.main_nf,
            )
        )
    else:
        self.passed.append(
            (
                "container_links",
                f"Correct number of double quotes found when specifying container: {line.lstrip('container ')}",
                self.main_nf,
            )
        )

    # Check for spaces in url
    single_quoted_items = raw_line.split("'")
    double_quoted_items = raw_line.split('"')
    # Look for container link as single item surrounded by quotes
    # (if there are multiple links, this will be warned in the next check)
    container_link = None
    if len(single_quoted_items) == 3:
        container_link = single_quoted_items[1]
    elif len(double_quoted_items) == 3:
        container_link = double_quoted_items[1]
    if container_link:
        if " " in container_link:
            self.failed.append(
                (
                    "container_links",
                    f"Space character found in container: '{container_link}'",
                    self.main_nf,
                )
            )
        else:
            self.passed.append(
                (
                    "container_links",
                    f"No space characters found in container: '{container_link}'",
                    self.main_nf,
                )
            )

        # lint more than one container in the same line
        if ("https://containers" in line or "https://depot" in line) and (
            "biocontainers/" in line or line.startswith(registry)
        ):
            self.warned.append(
                (
                    "container_links",
                    "Docker and Singularity containers specified in the same line. Only first one checked.",
                    self.main_nf,
                )
            )


def _parse_input(self, line_raw):
    """
    Return list of input channel names from an input line.

    If more than one elements in channel should work with both of:
        tuple val(meta), path(reads)
        tuple val(meta), path(reads, stageAs: "input*/*")

    If using a tuple, channel names must be in (parentheses)
    """
    inputs = []
    # Remove comments and trailing whitespace
    line = line_raw.split("//")[0]
    line = line.strip()
    # Tuples with multiple elements
    if "tuple" in line:
        matches = re.findall(r"\((\w+)\)", line)
        if matches:
            inputs.extend(matches)
        else:
            self.failed.append(
                (
                    "main_nf_input_tuple",
                    f"Found tuple but no channel names: `{line}`",
                    self.main_nf,
                )
            )
    # Single element inputs
    else:
        if "(" in line:
            match = re.search(r"\((\w+)\)", line)
            if match:
                inputs.append(match.group(1))
        else:
            inputs.append(line.split()[1])
    return inputs


def _parse_output(self, line):
    output = []
    if "meta" in line:
        output.append("meta")
    if "emit:" not in line:
        self.failed.append(("missing_emit", f"Missing emit statement: {line.strip()}", self.main_nf))
    else:
        output.append(line.split("emit:")[1].strip())

    return output


def _is_empty(line):
    """Check whether a line is empty or a comment"""
    empty = False
    if line.strip().startswith("//"):
        empty = True
    if line.strip().replace(" ", "") == "":
        empty = True
    return empty


def _fix_module_version(self, current_version, latest_version, singularity_tag, response):
    """Updates the module version

    Changes the bioconda current version by the latest version.
    Obtains the latest build from bioconda response
    Checks that the new URLs for docker and singularity with the tag [version]--[build] are valid
    Changes the docker and singularity URLs
    """
    # Get latest build
    build = _get_build(response)

    with open(self.main_nf) as source:
        lines = source.readlines()

    # Check if the new version + build exist and replace
    new_lines = []
    for line in lines:
        line_stripped = line.strip(" '\"")
        build_type = _container_type(line_stripped)
        if build_type == "conda":
            new_lines.append(re.sub(rf"{current_version}", f"{latest_version}", line))
        elif build_type in ("singularity", "docker"):
            # Check that the new url is valid
            new_url = re.search(
                "(?:['\"])(.+)(?:['\"])", re.sub(rf"{singularity_tag}", f"{latest_version}--{build}", line)
            ).group(1)
            try:
                response_new_container = requests.get(
                    "https://" + new_url if not new_url.startswith("https://") else new_url, stream=True
                )
                log.debug(
                    f"Connected to URL: {'https://' + new_url if not new_url.startswith('https://') else new_url}, "
                    f"status_code: {response_new_container.status_code}"
                )
            except (requests.exceptions.RequestException, sqlite3.InterfaceError) as e:
                log.debug(f"Unable to connect to url '{new_url}' due to error: {e}")
                return False
            if response_new_container.status_code != 200:
                return False
            new_lines.append(re.sub(rf"{singularity_tag}", f"{latest_version}--{build}", line))
        else:
            new_lines.append(line)

    # Replace outdated versions by the latest one
    with open(self.main_nf, "w") as source:
        for line in new_lines:
            source.write(line)

    return True


def _get_build(response):
    """Get the latest build of the container version"""
    build_times = []
    latest_v = response.get("latest_version")
    files = response.get("files")
    for f in files:
        if f.get("version") == latest_v:
            build_times.append((f.get("upload_time"), f.get("attrs").get("build")))
    return sorted(build_times, key=lambda tup: tup[0], reverse=True)[0][1]


def _container_type(line):
    """Returns the container type of a build."""
    if line.startswith("conda"):
        return "conda"
    if line.startswith("https://") or line.startswith("https://depot"):
        # Look for a http download URL.
        # Thanks Stack Overflow for the regex: https://stackoverflow.com/a/3809435/713980
        url_regex = (
            r"https?:\/\/(www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}\b([-a-zA-Z0-9()@:%_\+.~#?&//=]*)"
        )
        url_match = re.search(url_regex, line, re.S)
        if url_match:
            return "singularity"
        return None
    if line.count("/") >= 1 and line.count(":") == 1 and line.count(" ") == 0 and "https://" not in line:
        return "docker"



import logging
import mmap
import os
import shutil
from pathlib import Path
from typing import Dict, List, Optional, Union

import nf_core.utils
from nf_core.modules.modules_json import ModulesJson
from nf_core.modules.modules_repo import ModulesRepo

from .components_utils import get_repo_info

log = logging.getLogger(__name__)


class ComponentCommand:
    """
    Base class for the 'nf-core modules' and 'nf-core subworkflows' commands
    """

    def __init__(
        self,
        component_type: str,
        directory: Union[str, Path] = ".",
        remote_url: Optional[str] = None,
        branch: Optional[str] = None,
        no_pull: bool = False,
        hide_progress: bool = False,
        no_prompts: bool = False,
    ) -> None:
        """
        Initialise the ComponentClass object
        """
        self.component_type: str = component_type
        self.directory: Path = Path(directory)
        self.modules_repo = ModulesRepo(remote_url, branch, no_pull, hide_progress)
        self.hide_progress: bool = hide_progress
        self.no_prompts: bool = no_prompts
        self.repo_type: Optional[str] = None
        self.org: str = ""
        self._configure_repo_and_paths()

    def _configure_repo_and_paths(self, nf_dir_req: bool = True) -> None:
        """
        Determine the repo type and set some default paths.
        If this is a modules repo, determine the org_path too.

        Args:
            nf_dir_req (bool, optional): Whether this command requires being run in the nf-core modules repo or a nf-core pipeline repository. Defaults to True.
        """
        try:
            if self.directory:
                if self.directory == Path(".") and not nf_dir_req:
                    self.no_prompts = True
                self.directory, self.repo_type, self.org = get_repo_info(self.directory, use_prompt=not self.no_prompts)
        except UserWarning:
            if nf_dir_req:
                raise
        except FileNotFoundError:
            raise

        self.default_modules_path = Path("modules", self.org)
        self.default_tests_path = Path("tests", "modules", self.org)
        self.default_subworkflows_path = Path("subworkflows", self.org)
        self.default_subworkflows_tests_path = Path("tests", "subworkflows", self.org)

    def get_local_components(self) -> List[str]:
        """
        Get the local modules/subworkflows in a pipeline
        """
        local_component_dir = Path(self.directory, self.component_type, "local")
        return [
            str(path.relative_to(local_component_dir)) for path in local_component_dir.iterdir() if path.suffix == ".nf"
        ]

    def get_components_clone_modules(self) -> List[str]:
        """
        Get the modules/subworkflows repository available in a clone of nf-core/modules
        """
        if self.component_type == "modules":
            component_base_path = Path(self.directory, self.default_modules_path)
        elif self.component_type == "subworkflows":
            component_base_path = Path(self.directory, self.default_subworkflows_path)
        return [
            str(Path(directory).relative_to(component_base_path))
            for directory, _, files in os.walk(component_base_path)
            if "main.nf" in files
        ]

    def has_valid_directory(self) -> bool:
        """Check that we were given a pipeline or clone of nf-core/modules"""
        if self.repo_type == "modules":
            return True
        if not self.directory.exists():
            log.error(f"Could not find directory: {self.directory}")
            return False
        main_nf = Path(self.directory, "main.nf")
        nf_config = Path(self.directory, "nextflow.config")
        if not main_nf.exists() and not nf_config.exists():
            if self.directory.resolve().parts[-1].startswith("nf-core"):
                raise UserWarning(f"Could not find a 'main.nf' or 'nextflow.config' file in '{self.directory}'")
            log.warning(f"Could not find a 'main.nf' or 'nextflow.config' file in '{self.directory}'")
        return True

    def has_modules_file(self) -> None:
        """Checks whether a module.json file has been created and creates one if it is missing"""
        modules_json_path = Path(self.directory, "modules.json")
        if not modules_json_path.exists():
            log.info("Creating missing 'module.json' file.")
            ModulesJson(self.directory).create()

    def clear_component_dir(self, component_name: str, component_dir: Union[str, Path]) -> bool:
        """
        Removes all files in the module/subworkflow directory

        Args:
            component_name (str): The name of the module/subworkflow
            component_dir (str, Path): The path to the module/subworkflow

        """

        try:
            shutil.rmtree(component_dir)
            # remove all empty directories
            for dir_path, dir_names, filenames in os.walk(self.directory, topdown=False):
                if not dir_names and not filenames:
                    try:
                        Path(dir_path).rmdir()
                    except OSError:
                        pass
                    else:
                        log.debug(f"Deleted  directory: '{dir_path}'")

            log.debug(f"Successfully removed {self.component_type[:-1]} {component_name}")
            return True
        except OSError as e:
            log.error(f"Could not remove {self.component_type[:-1]} {component_name}: {e}")
            return False

    def components_from_repo(self, install_dir: str) -> List[str]:
        """
        Gets the modules/subworkflows installed from a certain repository

        Args:
            install_dir (str): The name of the directory where modules/subworkflows are installed

        Returns:
            [str]: The names of the modules/subworkflows
        """
        repo_dir = Path(self.directory, self.component_type, install_dir)
        if not repo_dir.exists():
            raise LookupError(f"Nothing installed from {install_dir} in pipeline")

        return [
            str(Path(dir_path).relative_to(repo_dir)) for dir_path, _, files in os.walk(repo_dir) if "main.nf" in files
        ]

    def install_component_files(
        self, component_name: str, component_version: str, modules_repo: ModulesRepo, install_dir: Union[str, Path]
    ) -> bool:
        """
        Installs a module/subworkflow into the given directory

        Args:
            component_name (str): The name of the module/subworkflow
            component_version (str): Git SHA for the version of the module/subworkflow to be installed
            modules_repo (ModulesRepo): A correctly configured ModulesRepo object
            install_dir (str, Path): The path to where the module/subworkflow should be installed (should be the 'modules/' or 'subworkflows/' dir of the pipeline)

        Returns:
            (bool): Whether the operation was successful of not
        """
        return modules_repo.install_component(component_name, install_dir, component_version, self.component_type)

    def load_lint_config(self) -> None:
        """Parse a pipeline lint config file.

        Load the '.nf-core.yml'  config file and extract
        the lint config from it

        Add parsed config to the `self.lint_config` class attribute.
        """
        _, tools_config = nf_core.utils.load_tools_config(self.directory)
        if tools_config is None:
            raise UserWarning("Could not load `.nf-core.yml` file.")
        else:
            self.lint_config = tools_config.get("lint", {})

    def check_modules_structure(self) -> None:
        """
        Check that the structure of the modules directory in a pipeline is the correct one:
            modules/nf-core/TOOL/SUBTOOL

        Prior to nf-core/tools release 2.6 the directory structure had an additional level of nesting:
            modules/nf-core/modules/TOOL/SUBTOOL
        """
        if self.repo_type == "pipeline":
            wrong_location_modules: List[Path] = []
            for directory, _, files in os.walk(Path(self.directory, "modules")):
                if "main.nf" in files:
                    module_path = Path(directory).relative_to(Path(self.directory, "modules"))
                    parts = module_path.parts
                    # Check that there are modules installed directly under the 'modules' directory
                    if parts[1] == "modules":
                        wrong_location_modules.append(module_path)
            # If there are modules installed in the wrong location
            if len(wrong_location_modules) > 0:
                log.info("The modules folder structure is outdated. Reinstalling modules.")
                # Remove the local copy of the modules repository
                log.info(f"Updating '{self.modules_repo.local_repo_dir}'")
                self.modules_repo.setup_local_repo(
                    self.modules_repo.remote_url, self.modules_repo.branch, self.hide_progress
                )
                # Move wrong modules to the right directory
                for module in wrong_location_modules:
                    modules_dir = Path("modules").resolve()
                    correct_dir = Path(modules_dir, self.modules_repo.repo_path, Path(*module.parts[2:]))
                    wrong_dir = Path(modules_dir, module)
                    shutil.move(str(wrong_dir), str(correct_dir))
                    log.info(f"Moved {wrong_dir} to {correct_dir}.")
                shutil.rmtree(Path(self.directory, "modules", self.modules_repo.repo_path, "modules"))
                # Regenerate modules.json file
                modules_json = ModulesJson(self.directory)
                modules_json.check_up_to_date()

    def check_patch_paths(self, patch_path: Path, module_name: str) -> None:
        """
        Check that paths in patch files are updated to the new modules path
        """
        if patch_path.exists():
            log.info(f"Modules {module_name} contains a patch file.")
            rewrite = False
            with open(patch_path) as fh:
                lines = fh.readlines()
                for index, line in enumerate(lines):
                    # Check if there are old paths in the patch file and replace
                    if f"modules/{self.modules_repo.repo_path}/modules/{module_name}/" in line:
                        rewrite = True
                        lines[index] = line.replace(
                            f"modules/{self.modules_repo.repo_path}/modules/{module_name}/",
                            f"modules/{self.modules_repo.repo_path}/{module_name}/",
                        )
            if rewrite:
                log.info(f"Updating paths in {patch_path}")
                with open(patch_path, "w") as fh:
                    for line in lines:
                        fh.write(line)
                # Update path in modules.json if the file is in the correct format
                modules_json = ModulesJson(self.directory)
                modules_json.load()
                if (
                    modules_json.has_git_url_and_modules()
                    and self.modules_repo.repo_path is not None
                    and modules_json.modules_json is not None
                ):
                    modules_json.modules_json["repos"][self.modules_repo.remote_url]["modules"][
                        self.modules_repo.repo_path
                    ][module_name]["patch"] = str(patch_path.relative_to(self.directory.resolve()))
                modules_json.dump()

    def check_if_in_include_stmts(self, component_path: str) -> Dict[str, List[Dict[str, Union[int, str]]]]:
        """
        Checks for include statements in the main.nf file of the pipeline and a list of line numbers where the component is included
        Args:
            component_path (str): The path to the module/subworkflow

        Returns:
            (list): A list of dictionaries, with the workflow file and the line number where the component is included
        """
        include_stmts: Dict[str, List[Dict[str, Union[int, str]]]] = {}
        if self.repo_type == "pipeline":
            workflow_files = Path(self.directory, "workflows").glob("*.nf")
            for workflow_file in workflow_files:
                with open(workflow_file) as fh:
                    # Check if component path is in the file using mmap
                    with mmap.mmap(fh.fileno(), 0, access=mmap.ACCESS_READ) as s:
                        if s.find(component_path.encode()) != -1:
                            # If the component path is in the file, check for include statements
                            for i, line in enumerate(fh):
                                if line.startswith("include") and component_path in line:
                                    if str(workflow_file) not in include_stmts:
                                        include_stmts[str(workflow_file)] = []
                                    include_stmts[str(workflow_file)].append(
                                        {"line_number": i + 1, "line": line.rstrip()}
                                    )

            return include_stmts
        else:
            log.debug("Not a pipeline repository, skipping check for include statements")
            return include_stmts



from pathlib import Path

import yaml

import nf_core.pipelines.lint

from ..test_lint import TestLint


class TestLintActionsAwsfulltest(TestLint):
    def test_actions_awsfulltest_warn(self):
        """Lint test: actions_awsfulltest - PASS"""
        self.lint_obj._load()
        results = self.lint_obj.actions_awsfulltest()
        assert "`.github/workflows/awsfulltest.yml` is triggered correctly" in results["passed"]
        assert len(results.get("failed", [])) == 0
        assert len(results.get("ignored", [])) == 0

    def test_actions_awsfulltest_pass(self):
        """Lint test: actions_awsfulltest - WARN"""

        # Edit .github/workflows/awsfulltest.yml to use -profile test_full
        new_pipeline = self._make_pipeline_copy()
        with open(Path(new_pipeline, ".github", "workflows", "awsfulltest.yml")) as fh:
            awsfulltest_yml = fh.read()
        awsfulltest_yml = awsfulltest_yml.replace("-profile test ", "-profile test_full ")
        with open(Path(new_pipeline, ".github", "workflows", "awsfulltest.yml"), "w") as fh:
            fh.write(awsfulltest_yml)

        # Make lint object
        lint_obj = nf_core.pipelines.lint.PipelineLint(new_pipeline)
        lint_obj._load()

        results = lint_obj.actions_awsfulltest()
        assert results["passed"] == [
            "`.github/workflows/awsfulltest.yml` is triggered correctly",
            "`.github/workflows/awsfulltest.yml` does not use `-profile test`",
        ]
        assert len(results.get("warned", [])) == 0
        assert len(results.get("failed", [])) == 0
        assert len(results.get("ignored", [])) == 0

    def test_actions_awsfulltest_fail(self):
        """Lint test: actions_awsfulltest - FAIL"""

        # Edit .github/workflows/awsfulltest.yml to use -profile test_full
        new_pipeline = self._make_pipeline_copy()
        with open(Path(new_pipeline, ".github", "workflows", "awsfulltest.yml")) as fh:
            awsfulltest_yml = yaml.safe_load(fh)
        del awsfulltest_yml[True]["pull_request_review"]
        with open(Path(new_pipeline, ".github", "workflows", "awsfulltest.yml"), "w") as fh:
            yaml.dump(awsfulltest_yml, fh)

        # Make lint object
        lint_obj = nf_core.pipelines.lint.PipelineLint(new_pipeline)
        lint_obj._load()

        results = lint_obj.actions_awsfulltest()
        assert results["failed"] == ["`.github/workflows/awsfulltest.yml` is not triggered correctly"]
        assert "`.github/workflows/awsfulltest.yml` does not use `-profile test`" in results["passed"]
        assert len(results.get("ignored", [])) == 0



"""
The NFCoreComponent class holds information and utility functions for a single module or subworkflow
"""

import logging
import re
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

log = logging.getLogger(__name__)


class NFCoreComponent:
    """
    A class to hold the information about a nf-core module or subworkflow.
    Includes functionality for linting.
    """

    def __init__(
        self,
        component_name: str,
        repo_url: Optional[str],
        component_dir: Path,
        repo_type: Optional[str],
        base_dir: Path,
        component_type: str,
        remote_component: bool = True,
    ):
        """
        Initialize the object

        Args:
            component_name (str): The name of the module or subworkflow
            repo_url (str): The URL of the repository
            component_dir (Path): The absolute path to the module or subworkflow
            repo_type (str): Either 'pipeline' or 'modules' depending on
                             whether the directory is a pipeline or clone
                             of nf-core/modules.
            base_dir (Path): The absolute path to the pipeline base dir
            component_type (str): Either 'modules' or 'subworkflows'
            remote_component (bool): Whether the module is to be treated as a
                                     nf-core or local component
        """
        self.component_type = component_type
        self.component_name = component_name
        self.repo_url = repo_url
        self.component_dir = component_dir
        self.repo_type = repo_type
        self.base_dir = base_dir
        self.passed: List[Tuple[str, str, Path]] = []
        self.warned: List[Tuple[str, str, Path]] = []
        self.failed: List[Tuple[str, str, Path]] = []
        self.inputs: List[List[Dict[str, Dict[str, str]]]] = []
        self.outputs: List[str] = []
        self.has_meta: bool = False
        self.git_sha: Optional[str] = None
        self.is_patched: bool = False
        self.branch: Optional[str] = None
        self.workflow_name: Optional[str] = None

        if remote_component:
            # Initialize the important files
            self.main_nf: Path = Path(self.component_dir, "main.nf")
            self.meta_yml: Optional[Path] = Path(self.component_dir, "meta.yml")
            self.environment_yml: Optional[Path] = Path(self.component_dir, "environment.yml")

            component_list = self.component_name.split("/")

            name_index = len(self.component_dir.parts) - 1 - self.component_dir.parts[::-1].index(component_list[0])
            if len(component_list) != 1 and component_list[0] == component_list[1]:
                # Handle cases where the subtool has the same name as the tool
                name_index -= 1

            repo_dir = self.component_dir.parts[:name_index][-1]

            self.org = repo_dir
            self.nftest_testdir = Path(self.component_dir, "tests")
            self.nftest_main_nf = Path(self.nftest_testdir, "main.nf.test")

            if self.repo_type == "pipeline":
                patch_fn = f"{self.component_name.replace('/', '-')}.diff"
                patch_path = Path(self.component_dir, patch_fn)
                if patch_path.exists():
                    self.is_patched = True
                    self.patch_path = patch_path
        else:
            # The main file is just the local module
            self.main_nf = self.component_dir
            self.component_name = self.component_dir.stem
            # These attributes are only used by nf-core modules
            # so just initialize them to None
            self.meta_yml = None
            self.environment_yml = None
            self.test_dir = None
            self.test_yml = None
            self.test_main_nf = None

        self.process_name: str = self._get_process_name()

    def __repr__(self) -> str:
        return f"<NFCoreComponent {self.component_name} {self.component_dir} {self.repo_url}>"

    def _get_main_nf_tags(self, test_main_nf: Union[Path, str]):
        """Collect all tags from the main.nf.test file."""
        tags = []
        with open(test_main_nf) as fh:
            for line in fh:
                if line.strip().startswith("tag"):
                    tags.append(line.strip().split()[1].strip('"'))
        return tags

    def _get_included_components(self, main_nf: Union[Path, str]):
        """Collect all included components from the main.nf file."""
        included_components = []
        with open(main_nf) as fh:
            for line in fh:
                if line.strip().startswith("include"):
                    # get tool/subtool or subworkflow name from include statement, can be in the form
                    #'../../../modules/nf-core/hisat2/align/main'
                    #'../bam_sort_stats_samtools/main'
                    #'../subworkflows/nf-core/bam_sort_stats_samtools/main'
                    #'plugin/nf-validation'
                    component = line.strip().split()[-1].split(self.org)[-1].split("main")[0].strip("/")
                    component = component.replace("'../", "subworkflows/")
                    component = component.replace("'", "")
                    included_components.append(component)
        return included_components

    def _get_included_components_in_chained_tests(self, main_nf_test: Union[Path, str]):
        """Collect all included components from the main.nf file."""
        included_components = []
        with open(main_nf_test) as fh:
            for line in fh:
                if line.strip().startswith("script"):
                    # get tool/subtool or subworkflow name from script statement, can be:
                    # if the component is a module TOOL/SUBTOOL:
                    # '../../SUBTOOL/main.nf'
                    # '../../../TOOL/SUBTOOL/main.nf'
                    # '../../../TOOL/main.nf'
                    # if the component is a module TOOL:
                    # '../../TOOL/main.nf'
                    # '../../TOOL/SUBTOOL/main.nf'
                    # if the component uses full paths or is a subworkflow:
                    # '(../../)modules/nf-core/TOOL/(SUBTOOL/)main.nf'
                    # '(../../)subworkflows/nf-core/TOOL/(SUBTOOL/)main.nf'
                    # the line which uses the current component script:
                    # '../main.nf'
                    component = (
                        line.strip()
                        .split("../")[-1]
                        .split(self.org)[-1]
                        .split("main.nf")[0]
                        .strip("'")
                        .strip('"')
                        .strip("/")
                    )
                    if (
                        "/" in self.component_name
                        and "/" not in component
                        and line.count("../") == 2
                        and self.org not in line
                        and component != ""
                    ):
                        # Add the current component name "TOOL" to the tag
                        component = f"{self.component_name.split('/')[0]}/{component}"
                    if "subworkflows" in line:
                        # Add the subworkflows prefix to the tag
                        component = f"subworkflows/{component}"
                    if component != "":
                        included_components.append(component)
        return included_components

    def _get_process_name(self):
        with open(self.main_nf) as fh:
            for line in fh:
                if re.search(r"^\s*process\s*\w*\s*{", line):
                    return re.search(r"^\s*process\s*(\w*)\s*{.*", line).group(1) or ""
        return ""

    def get_inputs_from_main_nf(self) -> None:
        """Collect all inputs from the main.nf file."""
        inputs: Any = []  # Can be 'list[list[dict[str, dict[str, str]]]]' or 'list[str]'
        with open(self.main_nf) as f:
            data = f.read()
        if self.component_type == "modules":
            # get input values from main.nf after "input:", which can be formatted as tuple val(foo) path(bar) or val foo or val bar or path bar or path foo
            # regex matches:
            # val(foo)
            # path(bar)
            # val foo
            # val bar
            # path bar
            # path foo
            # don't match anything inside comments or after "output:"
            if "input:" not in data:
                log.debug(f"Could not find any inputs in {self.main_nf}")
                return
            input_data = data.split("input:")[1].split("output:")[0]
            for line in input_data.split("\n"):
                channel_elements: Any = []
                regex = r"(val|path)\s*(\(([^)]+)\)|\s*([^)\s,]+))"
                matches = re.finditer(regex, line)
                for _, match in enumerate(matches, start=1):
                    input_val = None
                    if match.group(3):
                        input_val = match.group(3).split(",")[0]  # handle `files, stageAs: "inputs/*"` cases
                    elif match.group(4):
                        input_val = match.group(4).split(",")[0]  # handle `files, stageAs: "inputs/*"` cases
                    if input_val:
                        channel_elements.append({input_val: {}})
                if len(channel_elements) > 0:
                    inputs.append(channel_elements)
            log.debug(f"Found {len(inputs)} inputs in {self.main_nf}")
            self.inputs = inputs
        elif self.component_type == "subworkflows":
            # get input values from main.nf after "take:"
            if "take:" not in data:
                log.debug(f"Could not find any inputs in {self.main_nf}")
                return
            # get all lines between "take" and "main" or "emit"
            input_data = data.split("take:")[1].split("main:")[0].split("emit:")[0]
            for line in input_data.split("\n"):
                try:
                    inputs.append(line.split()[0])
                except IndexError:
                    pass  # Empty lines
            log.debug(f"Found {len(inputs)} inputs in {self.main_nf}")
            self.inputs = inputs

    def get_outputs_from_main_nf(self):
        outputs = []
        with open(self.main_nf) as f:
            data = f.read()
        if self.component_type == "modules":
            # get output values from main.nf after "output:". the names are always after "emit:"
            if "output:" not in data:
                log.debug(f"Could not find any outputs in {self.main_nf}")
                return outputs
            output_data = data.split("output:")[1].split("when:")[0]
            regex_emit = r"emit:\s*([^)\s,]+)"
            regex_elements = r"(val|path|env|stdout)\s*(\(([^)]+)\)|\s*([^)\s,]+))"
            for line in output_data.split("\n"):
                match_emit = re.search(regex_emit, line)
                matches_elements = re.finditer(regex_elements, line)
                if not match_emit:
                    continue
                output_channel = {match_emit.group(1): []}
                for _, match_element in enumerate(matches_elements, start=1):
                    output_val = None
                    if match_element.group(3):
                        output_val = match_element.group(3)
                    elif match_element.group(4):
                        output_val = match_element.group(4)
                    if output_val:
                        output_val = output_val.strip("'").strip('"')  # remove quotes
                        output_channel[match_emit.group(1)].append({output_val: {}})
                outputs.append(output_channel)
            log.debug(f"Found {len(outputs)} outputs in {self.main_nf}")
            self.outputs = outputs
        elif self.component_type == "subworkflows":
            # get output values from main.nf after "emit:". Can be named outputs or not.
            if "emit:" not in data:
                log.debug(f"Could not find any outputs in {self.main_nf}")
                return outputs
            output_data = data.split("emit:")[1].split("}")[0]
            for line in output_data.split("\n"):
                try:
                    outputs.append(line.split("=")[0].split()[0])
                except IndexError:
                    # Empty lines
                    pass
            log.debug(f"Found {len(outputs)} outputs in {self.main_nf}")
            self.outputs = outputs



import copy
import json
import shutil
from pathlib import Path

from nf_core.components.components_utils import (
    NF_CORE_MODULES_DEFAULT_BRANCH,
    NF_CORE_MODULES_NAME,
    NF_CORE_MODULES_REMOTE,
)
from nf_core.modules.modules_json import ModulesJson
from nf_core.modules.modules_repo import ModulesRepo
from nf_core.modules.patch import ModulePatch

from ..test_modules import TestModules


class TestModulesCreate(TestModules):
    def test_get_modules_json(self):
        """Checks that the get_modules_json function returns the correct result"""
        mod_json_path = Path(self.pipeline_dir, "modules.json")
        with open(mod_json_path) as fh:
            try:
                mod_json_sb = json.load(fh)
            except json.JSONDecodeError as e:
                raise UserWarning(f"Unable to load JSON file '{mod_json_path}' due to error {e}")

        mod_json_obj = ModulesJson(self.pipeline_dir)
        mod_json = mod_json_obj.get_modules_json()

        # Check that the modules.json hasn't changed
        assert mod_json == mod_json_sb

    def test_mod_json_update(self):
        """Checks whether the update function works properly"""
        mod_json_obj = ModulesJson(self.pipeline_dir)
        # Update the modules.json file
        mod_repo_obj = ModulesRepo()
        mod_json_obj.update("modules", mod_repo_obj, "MODULE_NAME", "GIT_SHA", ["modules"], write_file=False)
        mod_json = mod_json_obj.get_modules_json()
        assert "MODULE_NAME" in mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"]["nf-core"]
        assert "git_sha" in mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"]["nf-core"]["MODULE_NAME"]
        assert "GIT_SHA" == mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"]["nf-core"]["MODULE_NAME"]["git_sha"]
        assert (
            NF_CORE_MODULES_DEFAULT_BRANCH
            == mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"]["nf-core"]["MODULE_NAME"]["branch"]
        )
        assert (
            "modules" in mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"]["nf-core"]["MODULE_NAME"]["installed_by"]
        )

    def test_mod_json_create(self):
        """Test creating a modules.json file from scratch"""
        mod_json_path = Path(self.pipeline_dir, "modules.json")
        # Remove the existing modules.json file
        mod_json_path.unlink()

        # Create the new modules.json file
        # (There are no prompts as long as there are only nf-core modules)
        ModulesJson(self.pipeline_dir).create()

        # Check that the file exists
        assert (mod_json_path).exists()

        # Get the contents of the file
        mod_json_obj = ModulesJson(self.pipeline_dir)
        mod_json = mod_json_obj.get_modules_json()

        mods = ["fastqc", "multiqc"]
        for mod in mods:
            assert mod in mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"]["nf-core"]
            assert "git_sha" in mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"]["nf-core"][mod]
            assert "branch" in mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"]["nf-core"][mod]

    def _modify_main_nf(self, path):
        """Modify a file to test patch creation"""
        with open(path) as fh:
            lines = fh.readlines()
        # Modify $meta.id to $meta.single_end
        lines[1] = '    tag "$meta.single_end"\n'
        with open(path, "w") as fh:
            fh.writelines(lines)

    def test_mod_json_create_with_patch(self):
        """Test creating a modules.json file from scratch when there are patched modules"""
        mod_json_path = Path(self.pipeline_dir, "modules.json")

        # Modify the module
        module_path = Path(self.pipeline_dir, "modules", "nf-core", "fastqc")
        self._modify_main_nf(module_path / "main.nf")

        # Try creating a patch file
        patch_obj = ModulePatch(self.pipeline_dir, NF_CORE_MODULES_REMOTE, NF_CORE_MODULES_DEFAULT_BRANCH)
        patch_obj.patch("fastqc")

        # Remove the existing modules.json file
        mod_json_path.unlink()

        # Create the new modules.json file
        ModulesJson(self.pipeline_dir).create()

        # Check that the file exists
        assert mod_json_path.is_file()

        # Get the contents of the file
        mod_json_obj = ModulesJson(self.pipeline_dir)
        mod_json = mod_json_obj.get_modules_json()

        # Check that fastqc is in the file
        assert "fastqc" in mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"]["nf-core"]
        assert "git_sha" in mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"]["nf-core"]["fastqc"]
        assert "branch" in mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"]["nf-core"]["fastqc"]

        # Check that fastqc/main.nf maintains the changes
        with open(module_path / "main.nf") as fh:
            lines = fh.readlines()
        assert lines[1] == '    tag "$meta.single_end"\n'

    def test_mod_json_up_to_date(self):
        """
        Checks if the modules.json file is up to date
        when no changes have been made to the pipeline
        """
        mod_json_obj = ModulesJson(self.pipeline_dir)
        mod_json_before = mod_json_obj.get_modules_json()
        mod_json_obj.check_up_to_date()
        mod_json_after = mod_json_obj.get_modules_json()

        # Check that the modules.json hasn't changed
        assert mod_json_before == mod_json_after

    def test_mod_json_up_to_date_module_removed(self):
        """
        Reinstall a module that has an entry in the modules.json
        but is missing in the pipeline
        """
        # Remove the fastqc module
        fastqc_path = Path(self.pipeline_dir, "modules", NF_CORE_MODULES_NAME, "fastqc")
        shutil.rmtree(fastqc_path)

        # Check that the modules.json file is up to date, and reinstall the module
        mod_json_obj = ModulesJson(self.pipeline_dir)
        mod_json_obj.check_up_to_date()

        # Check that the module has been reinstalled
        files = ["main.nf", "meta.yml"]
        assert fastqc_path.exists()
        for f in files:
            assert Path(fastqc_path, f).exists()

    def test_mod_json_up_to_date_reinstall_fails(self):
        """
        Try reinstalling a module where the git_sha is invalid
        """
        mod_json_obj = ModulesJson(self.pipeline_dir)

        # Update the fastqc module entry to an invalid git_sha
        mod_json_obj.update("modules", ModulesRepo(), "fastqc", "INVALID_GIT_SHA", ["modules"], write_file=True)

        # Remove the fastqc module
        fastqc_path = Path(self.pipeline_dir, "modules", NF_CORE_MODULES_NAME, "fastqc")
        shutil.rmtree(fastqc_path)

        # Check that the modules.json file is up to date, and remove the fastqc module entry
        mod_json_obj.check_up_to_date()
        mod_json = mod_json_obj.get_modules_json()

        # Check that the module has been removed from the modules.json
        assert "fastqc" not in mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"]["nf-core"]

    def test_mod_json_repo_present(self):
        """Tests the repo_present function"""
        mod_json_obj = ModulesJson(self.pipeline_dir)

        assert mod_json_obj.repo_present(NF_CORE_MODULES_REMOTE) is True
        assert mod_json_obj.repo_present("INVALID_REPO") is False

    def test_mod_json_component_present(self):
        """Tests the component_present function"""
        mod_json_obj = ModulesJson(self.pipeline_dir)

        assert mod_json_obj.component_present("fastqc", NF_CORE_MODULES_REMOTE, NF_CORE_MODULES_NAME, "modules") is True
        assert (
            mod_json_obj.component_present("INVALID_MODULE", NF_CORE_MODULES_REMOTE, NF_CORE_MODULES_NAME, "modules")
            is False
        )
        assert mod_json_obj.component_present("fastqc", "INVALID_REPO", "INVALID_DIR", "modules") is False
        assert mod_json_obj.component_present("INVALID_MODULE", "INVALID_REPO", "INVALID_DIR", "modules") is False

    def test_mod_json_get_module_version(self):
        """Test the get_module_version function"""
        mod_json_obj = ModulesJson(self.pipeline_dir)
        mod_json = mod_json_obj.get_modules_json()
        assert (
            mod_json_obj.get_module_version("fastqc", NF_CORE_MODULES_REMOTE, NF_CORE_MODULES_NAME)
            == mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"]["nf-core"]["fastqc"]["git_sha"]
        )
        assert mod_json_obj.get_module_version("INVALID_MODULE", NF_CORE_MODULES_REMOTE, NF_CORE_MODULES_NAME) is None

    def test_mod_json_dump(self):
        """Tests the dump function"""
        mod_json_obj = ModulesJson(self.pipeline_dir)
        mod_json = mod_json_obj.get_modules_json()
        # Remove the modules.json file
        mod_json_path = Path(self.pipeline_dir, "modules.json")
        mod_json_path.unlink()

        # Check that the dump function creates the file
        mod_json_obj.dump()
        assert mod_json_path.exists()

        # Check that the dump function writes the correct content
        with open(mod_json_path) as f:
            try:
                mod_json_new = json.load(f)
            except json.JSONDecodeError as e:
                raise UserWarning(f"Unable to load JSON file '{mod_json_path}' due to error {e}")
        assert mod_json == mod_json_new

    def test_mod_json_with_empty_modules_value(self):
        # Load module.json and remove the modules entry
        mod_json_obj = ModulesJson(self.pipeline_dir)
        mod_json_obj.create()  # Create modules.json explicitly to get correct module sha
        mod_json_orig = mod_json_obj.get_modules_json()
        mod_json = copy.deepcopy(mod_json_orig)
        mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"] = {}
        # save the altered module.json and load it again to check if it will fix itself
        mod_json_obj.modules_json = mod_json
        mod_json_obj.dump()
        mod_json_obj_new = ModulesJson(self.pipeline_dir)
        mod_json_obj_new.check_up_to_date()
        mod_json_new = mod_json_obj_new.get_modules_json()
        assert mod_json_orig == mod_json_new

    def test_mod_json_with_missing_modules_entry(self):
        # Load module.json and remove the modules entry
        mod_json_obj = ModulesJson(self.pipeline_dir)
        mod_json_obj.create()  # Create modules.json explicitly to get correct module sha
        mod_json_orig = mod_json_obj.get_modules_json()
        mod_json = copy.deepcopy(mod_json_orig)
        mod_json["repos"][NF_CORE_MODULES_REMOTE].pop("modules")
        # save the altered module.json and load it again to check if it will fix itself
        mod_json_obj.modules_json = mod_json
        mod_json_obj.dump()
        mod_json_obj_new = ModulesJson(self.pipeline_dir)
        mod_json_obj_new.check_up_to_date()
        mod_json_new = mod_json_obj_new.get_modules_json()
        assert mod_json_orig == mod_json_new



import filecmp
import logging
import os
import shutil
from configparser import NoOptionError, NoSectionError
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Union

import git
from git.exc import GitCommandError

from nf_core.components.components_utils import (
    NF_CORE_MODULES_DEFAULT_BRANCH,
    NF_CORE_MODULES_NAME,
    NF_CORE_MODULES_REMOTE,
)
from nf_core.utils import load_tools_config

log = logging.getLogger(__name__)


class RemoteProgressbar(git.RemoteProgress):
    """
    An object to create a progressbar for when doing an operation with the remote.
    Note that an initialized rich Progress (progress bar) object must be passed
    during initialization.
    """

    def __init__(self, progress_bar, repo_name, remote_url, operation):
        """
        Initializes the object and adds a task to the progressbar passed as 'progress_bar'

        Args:
            progress_bar (rich.progress.Progress): A rich progress bar object
            repo_name (str): Name of the repository the operation is performed on
            remote_url (str): Git URL of the repository the operation is performed on
            operation (str): The operation performed on the repository, i.e. 'Pulling', 'Cloning' etc.
        """
        super().__init__()
        self.progress_bar = progress_bar
        self.tid = self.progress_bar.add_task(
            f"{operation} from [bold green]'{repo_name}'[/bold green] ([link={remote_url}]{remote_url}[/link])",
            start=False,
            state="Waiting for response",
        )

    def update(self, op_code, cur_count, max_count=None, message=""):
        """
        Overrides git.RemoteProgress.update.
        Called every time there is a change in the remote operation
        """
        if not self.progress_bar.tasks[self.tid].started:
            self.progress_bar.start_task(self.tid)
        if cur_count is not None and max_count is not None:
            cur_count = float(cur_count)
            max_count = float(max_count)
            state = f"{cur_count / max_count * 100:.1f}%"
        else:
            state = "Unknown"

        self.progress_bar.update(self.tid, total=max_count, completed=cur_count, state=state)


class SyncedRepo:
    """
    An object to store details about a locally cached code repository.
    """

    local_repo_statuses: Dict[str, bool] = {}
    no_pull_global = False

    @staticmethod
    def local_repo_synced(repo_name):
        """
        Checks whether a local repo has been cloned/pull in the current session
        """
        return SyncedRepo.local_repo_statuses.get(repo_name, False)

    @staticmethod
    def update_local_repo_status(repo_name, up_to_date):
        """
        Updates the clone/pull status of a local repo
        """
        SyncedRepo.local_repo_statuses[repo_name] = up_to_date

    @staticmethod
    def get_remote_branches(remote_url):
        """
        Get all branches from a remote repository

        Args:
            remote_url (str): The git url to the remote repository

        Returns:
            (set[str]): All branches found in the remote
        """
        try:
            unparsed_branches = git.Git().ls_remote(remote_url)
        except git.GitCommandError:
            raise LookupError(f"Was unable to fetch branches from '{remote_url}'")
        else:
            branches = {}
            for branch_info in unparsed_branches.split("\n"):
                sha, name = branch_info.split("\t")
                if name != "HEAD":
                    # The remote branches are shown as 'ref/head/branch'
                    branch_name = Path(name).stem
                    branches[sha] = branch_name
            return set(branches.values())

    def __init__(self, remote_url=None, branch=None, no_pull=False, hide_progress=False):
        """
        Initializes the object and clones the git repository if it is not already present
        """

        # This allows us to set this one time and then keep track of the user's choice
        SyncedRepo.no_pull_global |= no_pull

        # Check if the remote seems to be well formed
        if remote_url is None:
            remote_url = NF_CORE_MODULES_REMOTE

        self.remote_url = remote_url
        self.fullname = None
        self.local_repo_dir = None

        self.repo = None
        # TODO: SyncedRepo doesn't have this method and both the ModulesRepo and
        # the WorkflowRepo define their own including custom init methods. This needs
        # fixing.
        self.setup_local_repo(remote_url, branch, hide_progress)

        if self.local_repo_dir is None:
            raise ValueError("Repository not initialized")
        else:
            config_fn, repo_config = load_tools_config(self.local_repo_dir)
            if config_fn is not None and repo_config is not None:
                try:
                    self.repo_path = repo_config.org_path
                except KeyError:
                    raise UserWarning(f"'org_path' key not present in {config_fn.name}")

            # Verify that the repo seems to be correctly configured
            if self.repo_path != NF_CORE_MODULES_NAME or self.branch:
                self.verify_branch()

            # Convenience variable
            self.modules_dir = Path(self.local_repo_dir, "modules", self.repo_path)
            self.subworkflows_dir = Path(self.local_repo_dir, "subworkflows", self.repo_path)

            self.avail_module_names = None

    def __repr__(self) -> str:
        return f"SyncedRepo({self.remote_url}, {self.branch})"

    def setup_local_repo(self, remote_url, branch, hide_progress):
        pass

    def verify_sha(self, prompt, sha):
        """
        Verify that 'sha' and 'prompt' arguments are not provided together.
        Verify that the provided SHA exists in the repo.

        Arguments:
            prompt (bool):              prompt asking for SHA
            sha (str):                  provided sha
        """
        if prompt and sha is not None:
            log.error("Cannot use '--sha' and '--prompt' at the same time!")
            return False

        if sha:
            if not self.sha_exists_on_branch(sha):
                log.error(f"Commit SHA '{sha}' doesn't exist in '{self.remote_url}'")
                return False

        return True

    def setup_branch(self, branch):
        """
        Verify that we have a branch and otherwise use the default one.
        The branch is then checked out to verify that it exists in the repo.

        Args:
            branch (str): Name of branch
        """
        if branch is None:
            # Don't bother fetching default branch if we're using nf-core
            if self.remote_url == NF_CORE_MODULES_REMOTE:
                self.branch = NF_CORE_MODULES_DEFAULT_BRANCH
            else:
                self.branch = self.get_default_branch()
        else:
            self.branch = branch

        # Verify that the branch exists by checking it out
        self.branch_exists()

    def get_default_branch(self):
        """
        Gets the default branch for the repo (the branch origin/HEAD is pointing to)
        """
        origin_head = next(ref for ref in self.repo.refs if ref.name == "origin/HEAD")
        _, branch = origin_head.ref.name.split("/")
        return branch

    def branch_exists(self):
        """
        Verifies that the branch exists in the repository by trying to check it out
        """
        try:
            self.checkout_branch()
        except GitCommandError:
            raise LookupError(f"Branch '{self.branch}' not found in '{self.remote_url}'")

    def verify_branch(self):
        """
        Verifies the active branch conforms to the correct directory structure
        """
        dir_names = os.listdir(self.local_repo_dir)
        if "modules" not in dir_names:
            err_str = f"Repository '{self.remote_url}' ({self.branch}) does not contain the 'modules/' directory"
            if "software" in dir_names:
                err_str += (
                    ".\nAs of nf-core/tools version 2.0, the 'software/' directory should be renamed to 'modules/'"
                )
            raise LookupError(err_str)

    def checkout_branch(self):
        """
        Checks out the specified branch of the repository
        """
        try:
            self.repo.git.checkout(self.branch)
        except GitCommandError as e:
            if (
                self.fullname
                and "modules" in self.fullname
                and "Your local changes to the following files would be overwritten by checkout" in str(e)
            ):
                log.debug(f"Overwriting local changes in '{self.local_repo_dir}'")
                self.repo.git.checkout(self.branch, force=True)
            else:
                raise e

    def checkout(self, commit):
        """
        Checks out the repository at the requested commit

        Args:
            commit (str): Git SHA of the commit
        """
        try:
            self.repo.git.checkout(commit)
        except GitCommandError as e:
            if (
                self.fullname
                and "modules" in self.fullname
                and "Your local changes to the following files would be overwritten by checkout" in str(e)
            ):
                log.debug(f"Overwriting local changes in '{self.local_repo_dir}'")
                self.repo.git.checkout(self.branch, force=True)
            else:
                raise e

    def component_exists(self, component_name, component_type, checkout=True, commit=None):
        """
        Check if a module/subworkflow exists in the branch of the repo

        Args:
            component_name (str): The name of the module/subworkflow

        Returns:
            (bool): Whether the module/subworkflow exists in this branch of the repository
        """
        return component_name in self.get_avail_components(component_type, checkout=checkout, commit=commit)

    def get_component_dir(self, component_name: str, component_type: str) -> Path:
        """
        Returns the file path of a module/subworkflow directory in the repo.
        Does not verify that the path exists.
        Args:
            component_name (str): The name of the module/subworkflow

        Returns:
            component_path (str): The path of the module/subworkflow in the local copy of the repository
        """
        if component_type == "modules":
            return Path(self.modules_dir, component_name)
        elif component_type == "subworkflows":
            return Path(self.subworkflows_dir, component_name)
        else:
            raise ValueError(f"Invalid component type: {component_type}")

    def install_component(
        self, component_name: str, install_dir: Union[str, Path], commit: str, component_type: str
    ) -> bool:
        """
        Install the module/subworkflow files into a pipeline at the given commit

        Args:
            component_name (str): The name of the module/subworkflow
            install_dir (str): The path where the module/subworkflow should be installed
            commit (str): The git SHA for the version of the module/subworkflow to be installed
            component_type (str): Either 'modules' or 'subworkflows'

        Returns:
            (bool): Whether the operation was successful or not
        """
        # Check out the repository at the requested ref
        try:
            self.checkout(commit)
        except git.GitCommandError:
            return False

        # Check if the module/subworkflow exists in the branch
        if not self.component_exists(component_name, component_type, checkout=False):
            log.error(
                f"The requested {component_type[:-1]} does not exists in the branch '{self.branch}' of {self.remote_url}'"
            )
            return False

        # Copy the files from the repo to the install folder
        shutil.copytree(self.get_component_dir(component_name, component_type), Path(install_dir, component_name))

        # Switch back to the tip of the branch
        self.checkout_branch()
        return True

    def component_files_identical(self, component_name, base_path, commit, component_type):
        """
        Checks whether the module or subworkflow files in a pipeline are identical to the ones in the remote
        Args:
            component_name (str): The name of the module or subworkflow
            base_path (str): The path to the module/subworkflow in the pipeline

        Returns:
            (bool): Whether the pipeline files are identical to the repo files
        """
        if commit is None:
            self.checkout_branch()
        else:
            self.checkout(commit)
        component_files = ["main.nf", "meta.yml"]
        files_identical = {file: True for file in component_files}
        component_dir = self.get_component_dir(component_name, component_type)
        for file in component_files:
            try:
                files_identical[file] = filecmp.cmp(Path(component_dir, file), Path(base_path, file))
            except FileNotFoundError:
                log.debug(f"Could not open file: {Path(component_dir, file)}")
                continue
        self.checkout_branch()
        return files_identical

    def ensure_git_user_config(self, default_name: str, default_email: str) -> None:
        if self.repo is None:
            raise ValueError("Repository not initialized")
        try:
            with self.repo.config_reader() as git_config:
                user_name = git_config.get_value("user", "name", default=None)
                user_email = git_config.get_value("user", "email", default=None)
        except (NoOptionError, NoSectionError):
            user_name = user_email = None

        if not user_name or not user_email:
            with self.repo.config_writer() as git_config:
                if not user_name:
                    git_config.set_value("user", "name", default_name)
                if not user_email:
                    git_config.set_value("user", "email", default_email)

    def get_component_git_log(
        self, component_name: Union[str, Path], component_type: str, depth: Optional[int] = None
    ) -> Iterable[Dict[str, str]]:
        """
        Fetches the commit history the of requested module/subworkflow since a given date. The default value is
        not arbitrary - it is the last time the structure of the nf-core/modules repository was had an
        update breaking backwards compatibility.
        Args:
            component_name (str): Name of module/subworkflow
            modules_repo (SyncedRepo): A SyncedRepo object configured for the repository in question

        Returns:
            ( dict ): Iterator of commit SHAs and associated (truncated) message
        """
        if self.repo is None:
            raise ValueError("Repository not initialized")
        self.checkout_branch()
        component_path = Path(component_type, self.repo_path, component_name)

        commits_new_iter = self.repo.iter_commits(max_count=depth, paths=component_path)
        commits_old_iter = []
        if component_type == "modules":
            # Grab commits also from previous modules structure
            old_component_path = Path("modules", component_name)
            commits_old_iter = self.repo.iter_commits(max_count=depth, paths=old_component_path)

        try:
            commits_old = [{"git_sha": commit.hexsha, "trunc_message": commit.message} for commit in commits_old_iter]
            commits_new = [{"git_sha": commit.hexsha, "trunc_message": commit.message} for commit in commits_new_iter]
        except git.GitCommandError as e:
            log.error(
                f"Git error: {e}\n"
                "To solve this, you can try to remove the cloned rempository and run the command again.\n"
                f"This repository is typically found at `{self.local_repo_dir}`"
            )
            raise UserWarning
        commits = iter(commits_new + commits_old)

        return commits

    def get_latest_component_version(self, component_name, component_type):
        """
        Returns the latest commit in the repository
        """
        try:
            git_logs = list(self.get_component_git_log(component_name, component_type, depth=1))
            if not git_logs:
                return None
            return git_logs[0]["git_sha"]
        except Exception as e:
            log.debug(f"Could not get latest version of {component_name}: {e}")
            return None

    def sha_exists_on_branch(self, sha):
        """
        Verifies that a given commit sha exists on the branch
        """
        if self.repo is None:
            raise ValueError("Repository not initialized")
        self.checkout_branch()
        return sha in (commit.hexsha for commit in self.repo.iter_commits())

    def get_commit_info(self, sha):
        """
        Fetches metadata about the commit (dates, message, etc.)
        Args:
            commit_sha (str): The SHA of the requested commit
        Returns:
            message (str): The commit message for the requested commit
            date (str): The commit date for the requested commit
        Raises:
            LookupError: If the search for the commit fails
        """
        if self.repo is None:
            raise ValueError("Repository not initialized")
        self.checkout_branch()
        for commit in self.repo.iter_commits():
            if commit.hexsha == sha:
                message = commit.message.splitlines()[0]
                date_obj = commit.committed_datetime
                date = str(date_obj.date())
                return message, date
        raise LookupError(f"Commit '{sha}' not found in the '{self.remote_url}'")

    def get_avail_components(
        self, component_type: str, checkout: bool = True, commit: Optional[str] = None
    ) -> List[str]:
        """
        Gets the names of the modules/subworkflows in the repository. They are detected by
        checking which directories have a 'main.nf' file

        Returns:
            ([ str ]): The module/subworkflow names
        """
        if checkout:
            self.checkout_branch()
        if commit is not None:
            self.checkout(commit)
        # Get directory
        if component_type == "modules":
            directory = self.modules_dir
        elif component_type == "subworkflows":
            directory = self.subworkflows_dir
        # Module/Subworkflow directories are characterized by having a 'main.nf' file
        avail_component_names = [
            str(Path(dirpath).relative_to(directory)) for dirpath, _, files in os.walk(directory) if "main.nf" in files
        ]
        return avail_component_names

    def get_meta_yml(self, component_type, module_name):
        """
        Returns the contents of the 'meta.yml' file of a module

        Args:
            module_name (str): The name of the module

        Returns:
            (str): The contents of the file in text format
        """
        self.checkout_branch()
        if component_type == "modules":
            path = Path(self.modules_dir, module_name, "meta.yml")
        elif component_type == "subworkflows":
            path = Path(self.subworkflows_dir, module_name, "meta.yml")
        else:
            raise ValueError(f"Invalid component type: {component_type}")
        if not path.exists():
            return None
        with open(path) as fh:
            contents = fh.read()
        return contents
